from abc import ABC, abstractmethod
from typing import Callable

import numpy as np
import shortuuid
from tensorboardX import SummaryWriter

from harlow.surrogating.surrogate_model import Surrogate
from harlow.utils.metrics import rmse


class Sampler(ABC):
    def __init__(
        self,
        target_function: Callable[[np.ndarray], np.ndarray],
        surrogate_model: Surrogate,
        domain_lower_bound: np.ndarray,
        domain_upper_bound: np.ndarray,
        fit_points_x: np.ndarray = None,
        fit_points_y: np.ndarray = None,
        test_points_x: np.ndarray = None,
        test_points_y: np.ndarray = None,
        evaluation_metric: Callable = rmse,
        logging_metrics: list = None,
        verbose: bool = False,
        run_name: str = None,
        save_dir: str = "",
    ):
        self.domain_lower_bound = domain_lower_bound
        self.domain_upper_bound = domain_upper_bound
        self.target_function = target_function
        self.surrogate_model = surrogate_model
        self.fit_points_x = fit_points_x
        self.fit_points_y = fit_points_y
        self.test_points_x = test_points_x
        self.test_points_y = test_points_y
        self.evaluation_metric = evaluation_metric
        self.logging_metrics = (
            [self.evaluation_metric]
            if not logging_metrics
            else list(set(logging_metrics.append(self.evaluation_metric)))
        )
        self.verbose = verbose
        self.run_name = run_name
        self.save_dir = save_dir

        self.step_x = []
        self.step_y = []
        self.step_score = []
        self.step_iter = []
        self.step_fit_time = []
        self.step_gen_time = []

        if not run_name:
            self.run_name = Sampler.generate_run_name()
        # Init writer for live web-based logging.
        self.writer = SummaryWriter(comment="-" + self.run_name)

    @abstractmethod
    def sample(
        self,
        n_initial_points: int = 20,
        n_new_points_per_iteration: int = 1,
        stopping_criterium: float = 0.05,
        max_n_iterations: int = 5000,
    ):
        pass

    def observer(self, X):
        """
        Wrapper for the user-specified `target_function` that checks input
        and output consistency.
        """

        # Check input points `X`
        if not isinstance(X, np.ndarray):
            raise ValueError(
                f"Parameters `X` must be of type {np.ndarray} but are"
                f" of type {type(X)} "
            )
        if X.ndim != 2:
            raise ValueError(
                f"Input array `X` must have shape `(n_points, n_features)`"
                f" but has shape {X.shape}."
            )

        # Call the target function
        y = self.target_function(X)

        # Check target `y` is a numpy array
        if not isinstance(y, np.ndarray):
            raise ValueError(
                f"Targets `y` must be of type {np.ndarray} but are of"
                f" type {type(y)}."
            )

        # Check shape of `y`
        if y.ndim != 2:
            raise ValueError(
                f"Target array `y` must have exactly 2 dimensions and shape "
                f"(n_points, n_outputs) but has {y.ndim} dimensions and shape "
                f"{y.shape} "
            )

        # Check consistency of input and output shapes
        if y.shape[0] != X.shape[0]:
            raise ValueError(
                f"Size of input array `X` and output array `y` must match for "
                f"dimension 0 but are {X.shape[0]} and {y.shape[0]} respectively."
            )

        return y

    @staticmethod
    def generate_run_name():
        return shortuuid.uuid()



"""
Surrogate model unit tests, to ensure that all surrogates:

    * Correctly inherit from the `Surrogate` abstract class
    * Can be initialized with the same set of keyword arguments
    * Have consistent input and output shapes
    * Behave consistently when calling fit, update, predict
"""

import numpy as np

from harlow.surrogating.surrogate_model import (  # noqa F401; DeepKernelMultiTaskGaussianProcess,
    BatchIndependentGaussianProcess,
    BayesianNeuralNetwork,
    GaussianProcessTFP,
    ModelListGaussianProcess,
    MultiTaskGaussianProcess,
    NeuralNetwork,
    VanillaGaussianProcess,
)

surrogate_list = [
    BatchIndependentGaussianProcess(),
    BayesianNeuralNetwork(),
    # DeepKernelMultiTaskGaussianProcess(),
    GaussianProcessTFP(),
    ModelListGaussianProcess(),
    MultiTaskGaussianProcess(),
    NeuralNetwork(),
    VanillaGaussianProcess(),
]

N_train = 20
N_update = 10
N_pred = 50
N_features = 10
N_outputs = 1


train_X = np.random.rand(N_train, N_features)
train_y = np.random.rand(N_train, N_outputs)
update_X = np.random.rand(N_update, N_features)
update_y = np.random.rand(N_update, N_outputs)
pred_X = np.random.rand(N_pred, N_features)


def test_surrogate_single_output():

    for surrogate in surrogate_list:

        # Training step
        try:
            surrogate.fit(train_X, train_y)
        except Exception as e:
            raise Exception(f"Surrogate fit failed for {surrogate} with exception: {e}")

        # TODO: Check number of samples internally for each surrogate model to make
        #   sure that `update_X` and `update_y` are added to the existing samples.

        # Training step
        try:
            surrogate.update(update_X, update_y)
        except Exception as e:
            raise Exception(
                f"Surrogate update failed for {surrogate} with exception: {e}"
            )

        # Prediction step
        try:
            pred_y = surrogate.predict(pred_X)
        except Exception as e:
            raise Exception(
                f"Surrogate prediction failed for {surrogate} with exception: {e}"
            )

        print(f"Assertions for surrogate {surrogate}")
        assert isinstance(pred_y, np.ndarray)
        assert pred_y.ndim == 2
        assert pred_y.shape[0] == N_pred
        assert pred_y.shape[1] == N_outputs


# if __name__ == "__main__":
#     test_surrogate_single_output()



"""Weak tests: does the code run without an error? the figures should be checked by
a human."""

import numpy as np

from harlow.utils.test_functions import ackley_nd
from harlow.visualization.corner import corner


def func_nd(x: np.ndarray):
    x = np.atleast_2d(x)
    return np.sum(x**2, axis=1)


def test_1d():
    support_range = np.array([[0, 2]]).T
    corner(func=func_nd, support_range=support_range)


def test_2d():
    support_range = np.array([[0, 2], [-1, 1]]).T
    return corner(func=func_nd, support_range=support_range)


def test_3d():
    support_range = np.array([[-2, 2], [-1, 1], [-np.pi, np.pi]]).T
    dim_labels = ["$\\theta$", "$A$", "$\\alpha_\\mathrm{x}$"]

    def func(x: np.ndarray):
        x = np.atleast_2d(x)
        x1 = x[:, 0]
        x2 = x[:, 1]
        x3 = x[:, 2]
        return x1 + x2**2 + np.sin(x3)

    corner(func=func, support_range=support_range)
    corner(func=func, support_range=support_range, dim_labels=dim_labels, iso_value=0)


def test_3d_rp33():
    # https://rprepo.readthedocs.io/en/latest/reliability_problems.html#sec-rp-33
    support_range = np.array([[-4, 4], [-4, 4], [-4, 4]]).T

    def func(x: np.ndarray):
        x = np.atleast_2d(x)
        x1 = x[:, 0]
        x2 = x[:, 1]
        x3 = x[:, 2]
        g1 = -x1 - x2 - x3 + 3 * np.sqrt(3)
        g2 = -x3 + 3
        return np.amin(np.stack((g1, g2)), 0)

    corner(func=func, support_range=support_range, iso_value=0)


def test_ackley_nd():
    n_dim = 3
    support_range = np.tile(np.array([[-4], [4]]), reps=(1, n_dim))
    corner(func=ackley_nd, support_range=support_range, n_discr=100)



import math

import numpy as np
from sklearn.metrics import mean_squared_error

from harlow.sampling.lola_voronoi import LolaVoronoi
from harlow.surrogating.surrogate_model import NN
from harlow.utils.test_functions import shekel


def shekel_benchmark():
    domain = np.array([[0.0, 10.0], [0.0, 10.0], [0.0, 10.0], [0.0, 10.0]])
    n_points = 200

    X1 = np.random.uniform(domain[0, 0], domain[0, 1], n_points)
    X2 = np.random.uniform(domain[1, 0], domain[1, 1], n_points)
    X3 = np.random.uniform(domain[2, 0], domain[2, 1], n_points)
    X4 = np.random.uniform(domain[3, 0], domain[3, 1], n_points)

    X = np.stack([X1, X2, X3, X4], -1)

    train_X = X[0:10, :]
    test_X = X[1:, :]
    train_y = shekel(train_X)
    test_y = shekel(test_X)

    nn = NN()
    nn.create_model(input_dim=(4,))
    nn.fit(train_X, train_y)
    y_hat = nn.predict(test_X)
    rmse = math.sqrt(mean_squared_error(test_y, y_hat))

    print(f"RMSE with initial trainingset of size {train_X.shape[0]}: {rmse}")
    experiment = [10, 50, 100, 150, 200]

    results = run_lv_experiment(
        nn, train_X, train_y, test_X, test_y, domain, experiment
    )
    print(results)

    results_random = run_random_experiment(
        nn, shekel, train_X, train_y, test_X, test_y, domain, experiment
    )
    print(results_random)


def run_random_experiment(
    model, test_fun, train_X, train_y, test_X, test_y, domain, n_samples_list
):
    results = {}

    for i in n_samples_list:
        X1 = np.random.uniform(domain[0, 0], domain[0, 1], i)
        X2 = np.random.uniform(domain[1, 0], domain[1, 1], i)
        X3 = np.random.uniform(domain[2, 0], domain[2, 1], i)
        X4 = np.random.uniform(domain[3, 0], domain[3, 1], i)
        new_X = np.stack([X1, X2, X3, X4], -1)

        for j in range(0, i):
            new_y = test_fun(np.expand_dims(new_X[j, :], axis=0))
            model.update(np.expand_dims(new_X[j, :], axis=0), new_y)

        y_hat = model.predict(test_X)
        rmse = math.sqrt(mean_squared_error(test_y, y_hat))
        results[i] = rmse
        print(f"RMSE with trainingset of size {train_X.shape[0]+i}: {rmse}")

    return results


def run_lv_experiment(model, train_X, train_y, test_X, test_y, domain, n_samples_list):
    n_per_iter = 1
    results = {}
    for i in n_samples_list:
        print(f"Running LV with {i} iterations")
        lv = LolaVoronoi(
            model,
            train_X,
            train_y,
            test_X,
            test_y,
            domain,
            shekel,
            n_iteration=i,
            n_new_point_per_iteration=n_per_iter,
            evaluation_metric="rmse",
        )
        lv.run_sequential_design()
        y_hat = lv.surrogate_model.predict(test_X)
        rmse = math.sqrt(mean_squared_error(test_y, y_hat))
        results[i] = rmse
        print(f"RMSE with trainingset of size {lv.train_X.shape[0]}: {rmse}")

    return results



"""
Creates a dictionary with truck information for influence line calculation. Two
setups are included: the single sensor TNO measurement setup used initially
and the additional fugro sensors.

STRUCTURE:
    truck_data -> series -> lane -> data
"""

import numpy as np

# ============================================================================
# Assemble dictionaries
#
# Create a dictionary to store all the info. This will include the truck data
# for the two measurement series:
#   * TNO measurements from strain gauge at the second span
#   * Fugro measurements from multiple strain gauges
#
# STRUCTURE:
#     truck_data -> series -> lane -> data
# ============================================================================

keys_data = ["force", "axle_dist", "center_z", "wheel_z_dist"]
keys_lane = ["left", "right"]
keys_series = ["TNO", "fugro", "uniform"]

# Create the nested dictionary
truck_data = dict.fromkeys(keys_series)
for _i, key_i in enumerate(truck_data.keys()):
    truck_data[key_i] = dict.fromkeys(keys_lane)
    for _j, key_j in enumerate(truck_data[key_i]):
        truck_data[key_i][key_j] = dict.fromkeys(keys_data)

# ============================================================================
# TRUCK DATA
# ============================================================================

right_truck_force = np.array(
    [
        [57.518 / 2, 57.518 / 2],
        [105.45 / 2, 105.45 / 2],
        [105.45 / 2, 105.45 / 2],
        [105.45 / 2, 105.45 / 2],
        [105.45 / 2, 105.45 / 2],
    ]
)
right_truck_axle_dist_x = np.array([1.94, 2.09, 1.345, 1.25])
right_truck_center_z = 5.700 / 2 - 3.625 / 2
right_truck_wheel_z_dist = 2.1500
right_truck_num_axle = np.shape(right_truck_force)[0]

left_truck_force = np.array(
    [
        [58.86 / 2, 58.86 / 2],
        [107.91 / 2, 107.91 / 2],
        [107.91 / 2, 107.91 / 2],
        [107.91 / 2, 107.91 / 2],
        [107.91 / 2, 107.91 / 2],
    ]
)
left_truck_axle_dist_x = np.array([2.0, 1.82, 1.82, 1.82])
left_truck_center_z = 5.700 / 2 + 3.625 / 2
left_truck_wheel_z_dist = right_truck_wheel_z_dist
left_truck_num_axle = np.shape(left_truck_force)[0]

right_ax_flatlist = np.cumsum(right_truck_axle_dist_x)
right_ax_flatlist = np.insert(right_ax_flatlist, 0, 0)

right_truck_x = np.repeat(right_ax_flatlist, 2)
right_truck_z = np.hstack(
    (
        right_truck_center_z
        - right_truck_wheel_z_dist / 2 * np.ones(right_truck_num_axle),
        right_truck_center_z
        + right_truck_wheel_z_dist / 2 * np.ones(right_truck_num_axle),
    )
)

left_ax_flatlist = np.cumsum(left_truck_axle_dist_x)
left_ax_flatlist = np.insert(left_ax_flatlist, 0, 0)

left_truck_x = np.repeat(left_ax_flatlist, 2)
left_truck_z = np.hstack(
    (
        left_truck_center_z
        - left_truck_wheel_z_dist / 2 * np.ones(left_truck_num_axle),
        left_truck_center_z
        + left_truck_wheel_z_dist / 2 * np.ones(left_truck_num_axle),
    )
)

# Assuming that z = 0 at the center of the bridge
z_truck_r_wheel_r = right_truck_center_z - right_truck_wheel_z_dist / 2
z_truck_r_wheel_l = right_truck_center_z + right_truck_wheel_z_dist / 2
z_truck_l_wheel_r = left_truck_center_z - left_truck_wheel_z_dist / 2
z_truck_l_wheel_l = left_truck_center_z + left_truck_wheel_z_dist / 2

# Fugro truck forces: the total weight is given as 50.420 tonnes.
# The force is calculated as Wtot * g * 0.12 for the first axle and
# Wtot * g * 0.22 for the other axles
truck_axle_dist_x_fugro = np.array([2.06, 1.83, 1.82, 1.82])
truck_force_fugro = np.array(
    [
        [59.35 / 2, 59.35 / 2],
        [108.82 / 2, 108.82 / 2],
        [108.82 / 2, 108.82 / 2],
        [108.82 / 2, 108.82 / 2],
        [108.82 / 2, 108.82 / 2],
    ]
)


# ============================================================================
# Pass the values to the dictionary
# ============================================================================
truck_data["TNO"]["left"]["force"] = left_truck_force
truck_data["TNO"]["left"]["axle_dist"] = left_truck_axle_dist_x
truck_data["TNO"]["left"]["center_z"] = left_truck_center_z
truck_data["TNO"]["left"]["wheel_z_dist"] = left_truck_wheel_z_dist
truck_data["TNO"]["left"]["z_wheel_r"] = z_truck_l_wheel_r
truck_data["TNO"]["left"]["z_wheel_l"] = z_truck_l_wheel_l

truck_data["TNO"]["right"]["force"] = right_truck_force
truck_data["TNO"]["right"]["axle_dist"] = right_truck_axle_dist_x
truck_data["TNO"]["right"]["center_z"] = right_truck_center_z
truck_data["TNO"]["right"]["wheel_z_dist"] = right_truck_wheel_z_dist
truck_data["TNO"]["right"]["z_wheel_r"] = z_truck_r_wheel_r
truck_data["TNO"]["right"]["z_wheel_l"] = z_truck_r_wheel_l

truck_data["fugro"]["left"]["force"] = truck_force_fugro
truck_data["fugro"]["left"]["axle_dist"] = truck_axle_dist_x_fugro
truck_data["fugro"]["left"]["center_z"] = left_truck_center_z
truck_data["fugro"]["left"]["wheel_z_dist"] = left_truck_wheel_z_dist
truck_data["fugro"]["left"]["z_wheel_r"] = z_truck_l_wheel_r
truck_data["fugro"]["left"]["z_wheel_l"] = z_truck_l_wheel_l

truck_data["fugro"]["right"]["force"] = truck_force_fugro
truck_data["fugro"]["right"]["axle_dist"] = truck_axle_dist_x_fugro
truck_data["fugro"]["right"]["center_z"] = right_truck_center_z
truck_data["fugro"]["right"]["wheel_z_dist"] = left_truck_wheel_z_dist
truck_data["fugro"]["right"]["z_wheel_r"] = z_truck_r_wheel_r
truck_data["fugro"]["right"]["z_wheel_l"] = z_truck_r_wheel_l



"""Corner plot visualization of R^n -> R functions. Plotting interesting 1D and 2D
sections."""
import copy
from typing import Callable, List

import labellines
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Rectangle

from harlow import RANGE_FACECOLOR, REAL_TYPE, REAL_VECT_TYPE


def corner(
    func: Callable[[np.ndarray], np.ndarray],
    support_range: REAL_VECT_TYPE,
    title: str = "",
    n_discr: int = 40,
    fixed_var_in_support_range: REAL_TYPE = 0.5,
    func_label: str = "$f$",
    dim_labels: List[str] = None,
    iso_value: REAL_TYPE = None,
    rectangular_range: REAL_VECT_TYPE = None,
):
    """Corner plot of a function (`func`): interesting 1D and 2D sections.
    For each subplot, each not plotted independent variable (`x_i`) is fixed to a
    constant value: `x_i_fix = x_i_support_lower_bound + fixed_var_in_support_range *
    x_i_support_range`.
    """
    # .....................................
    # Initialize
    # .....................................
    n_dim = support_range.shape[1]
    fixed_vars = support_range[0, :] + fixed_var_in_support_range * np.diff(
        support_range, axis=0
    )

    z_mx_off_diag_all = np.empty((n_discr, n_discr, n_dim, n_dim))
    z_mx_off_diag_all[:] = np.nan
    x_mx_off_diag_all = copy.deepcopy(z_mx_off_diag_all)
    y_mx_off_diag_all = copy.deepcopy(z_mx_off_diag_all)
    z_mx_diag_all = np.empty((n_discr, n_dim, n_dim))
    z_mx_diag_all[:] = np.nan

    if dim_labels is None:
        dim_labels = [f"$x_{{{ii+1}}}$" for ii in range(n_dim)]

    # .....................................
    # Plot
    # .....................................
    fig, axes = plt.subplots(n_dim, n_dim, figsize=(n_dim + 3, n_dim + 3))

    if n_dim == 1:
        axes = np.atleast_2d(axes)

    for row in range(n_dim):
        for col in range(n_dim):
            # turn off the upper triangle
            if row < col:
                axes[row, col].axis("off")

            # diagonal
            elif row == col:
                ax = axes[row, col]

                # print(f"row={row+1}/{n_dim}; \t col={col+1}/{n_dim} \t (diagonal)")
                y_labelpad = 15

                x_vec = np.linspace(
                    support_range[0, col], support_range[1, col], n_discr
                )
                all_x_vec = np.tile(fixed_vars, (n_discr, 1))

                all_x_vec[:, col] = x_vec
                z_vec = func(all_x_vec).ravel()

                z_mx_diag_all[:, row, col] = z_vec

                axes[col, col].plot(x_vec, z_vec, color="black")

                if iso_value is not None:
                    lh = ax.plot(
                        support_range[:, col], [iso_value] * 2, ls="--", color="red"
                    )
                    labellines.labelLines(lh, fontsize=7, outline_width=3)

                if rectangular_range is not None:
                    ax.axvspan(
                        rectangular_range[0, col],
                        rectangular_range[1, col],
                        alpha=0.2,
                        facecolor=RANGE_FACECOLOR,
                    )

                # left top corner
                if col == 0:
                    if n_dim == 1:
                        ax.set_xlabel(dim_labels[col])
                    else:
                        ax.set_xticks([])
                    ax.yaxis.tick_right()
                    ax.yaxis.set_label_position("right")
                    ax.set_ylabel(func_label, rotation=0, labelpad=y_labelpad)

                # right bottom corner
                elif col == n_dim - 1:
                    ax.set_xlabel(dim_labels[-1])
                    ax.yaxis.tick_right()
                    ax.yaxis.set_label_position("right")
                    ax.set_ylabel(func_label, rotation=0, labelpad=y_labelpad)

                # the rest
                else:
                    ax.set_xticks([])
                    ax.yaxis.tick_right()
                    ax.yaxis.set_label_position("right")
                    ax.set_ylabel(func_label, rotation=0, labelpad=y_labelpad)

            # off diagonal
            else:
                # print(f"row={row+1}/{n_dim}; \t col={col+1}/{n_dim} \t (off-diagonal)")  # noqa E501
                y_labelpad = 10

                x_vec = np.linspace(
                    support_range[0, col], support_range[1, col], n_discr
                )
                y_vec = np.linspace(
                    support_range[0, row], support_range[1, row], n_discr
                )
                x_mx, y_mx = np.meshgrid(x_vec, y_vec)

                all_x_vec = np.tile(fixed_vars, (n_discr * n_discr, 1))
                all_x_vec[:, col] = x_mx.ravel()
                all_x_vec[:, row] = y_mx.ravel()

                z_vec = func(all_x_vec)
                z_mx = z_vec.reshape((n_discr, n_discr))

                z_mx_off_diag_all[:, :, row, col] = z_mx
                x_mx_off_diag_all[:, :, row, col] = x_mx
                y_mx_off_diag_all[:, :, row, col] = y_mx

                # left bottom corner
                if row == n_dim - 1 and col == 0:
                    axes[row, col].set_xlabel(dim_labels[0])
                    axes[row, col].set_ylabel(
                        dim_labels[-1], rotation=0, labelpad=y_labelpad
                    )

                # bottom edge x label
                elif row == n_dim - 1:
                    if col != 0:
                        axes[row, col].set_yticks([])
                    axes[row, col].set_xlabel(dim_labels[col])

                # left edge x label
                elif col == 0:
                    axes[row, col].set_xticks([])
                    axes[row, col].set_ylabel(
                        dim_labels[row], rotation=0, labelpad=y_labelpad
                    )

                # rest
                else:
                    axes[row, col].set_yticks([])
                    axes[row, col].set_xticks([])

            # Format the axes - for all plots
            axes[row, col].tick_params(labelsize=6)
            axes[row, col].xaxis.offsetText.set_fontsize(6)
            axes[row, col].yaxis.offsetText.set_fontsize(6)
            axes[row, col].ticklabel_format(scilimits=(-5, 5))

    # ----------------------------------------------------------------------------------
    # SET THE RANGES AND COLORS
    # ----------------------------------------------------------------------------------
    # get the range of the ordinate
    range_tot = [
        np.nanmin([np.nanmin(z_mx_diag_all), np.nanmin(z_mx_off_diag_all)]),
        np.nanmax([np.nanmax(z_mx_diag_all), np.nanmax(z_mx_off_diag_all)]),
    ]

    levels = np.linspace(
        signif_floor(range_tot[0]), signif_ceil(range_tot[1]), 10
    ).ravel()
    # padding
    r = range_tot[1] - range_tot[0]
    d = 0.05
    range_tot = [range_tot[0] - d * r, range_tot[1] + d * r]

    # Set axis ranges and colors
    for row in range(n_dim):
        for col in range(n_dim):
            ax = axes[row, col]
            # upper triangle
            if row < col:
                pass  # do nothing

            # diagonal
            elif row == col:
                ax.set_ylim([range_tot[0], range_tot[1]])

            # off diagonal
            else:
                x_mx = x_mx_off_diag_all[:, :, row, col]
                y_mx = y_mx_off_diag_all[:, :, row, col]
                z_mx = z_mx_off_diag_all[:, :, row, col]

                color = ax.contourf(
                    x_mx,
                    y_mx,
                    z_mx,
                    levels=levels,
                )

                # data range
                if rectangular_range is not None:
                    ax.add_patch(
                        Rectangle(
                            xy=(rectangular_range[0, col], rectangular_range[0, row]),
                            width=np.diff(rectangular_range[:, col])[0],
                            height=np.diff(rectangular_range[:, row])[0],
                            facecolor=RANGE_FACECOLOR,
                            edgecolor=RANGE_FACECOLOR,
                            ls="--",
                            alpha=0.3,
                        )
                    )

                # Add iso line at level `iso_value`
                if iso_value is not None:
                    iso_lines = iso_line_2d(
                        x_mx=x_mx,
                        y_mx=y_mx,
                        z_mx=z_mx,
                        iso_level=iso_value,
                    )
                    for iso_line in iso_lines:
                        lh = ax.plot(
                            iso_line[:, 0],
                            iso_line[:, 1],
                            ls="--",
                            color="red",
                            label=f"{iso_value:.2f}",
                        )
                        try:
                            labellines.labelLines(
                                lh, fontsize=6, backgroundcolor="none", outline_width=3
                            )
                        except Exception as e:
                            if str(e) == "x label location is outside data range!":
                                pass
                            else:
                                raise e

    # ----------------------------------------------------------------------------------
    # COLORBAR
    # ----------------------------------------------------------------------------------
    if n_dim > 1:
        y_labelpad = 10
        cax = plt.axes([0.85, 0.55, 0.025, 0.3])
        cb = fig.colorbar(color, cax=cax)
        cb.ax.set_ylabel(func_label, rotation=0, labelpad=y_labelpad)
        cb.ax.tick_params(labelsize=6)
        cb.ax.yaxis.get_offset_text().set_fontsize(6)
        cb.formatter.set_powerlimits((-5, 5))
        cb.update_ticks()
        if iso_value is not None:
            cb.ax.plot(
                [np.min(range_tot), np.max(range_tot)],
                [iso_value] * 2,
                ls="--",
                color="red",
            )
    else:
        plt.tight_layout()

    plt.suptitle(title)

    return fig, axes


def iso_line_2d(x_mx, y_mx, z_mx, iso_level):
    """Get coordinates for a 2d contour plot.
    Args:
        x_mx: grid coordinates along the x-axis.
        y_mx: grid coordinates along the y-axis.
        z_mx: grid coordinates along the z-axis.
        iso_level: z value for which the contour plot is obtained.
    Returns:
        p: array of coordinates of the credible region.
    """
    plt.ioff()
    fig_tmp, ax_tmp = plt.subplots()
    cs = ax_tmp.contour(x_mx, y_mx, z_mx, levels=[iso_level])

    p = []
    for collection in cs.collections:
        for path in collection.get_paths():
            p.append(path.vertices)
    plt.close(fig_tmp)
    plt.ion()
    return p


def signif_ceil(x, digits: int = 3):
    """
    x = sci_coeff * 10**sci_exp
    """
    x = np.atleast_1d(x)
    idx = np.abs(x) < np.finfo(float).eps
    x[idx] = 3 * np.finfo(float).eps
    # scientific notation
    sci_exp = np.array([int(np.floor(np.log10(np.abs(x))))])
    sci_scale = 10.0**sci_exp
    sci_coeff = x / sci_scale

    # number of digits
    dig_scale = 10.0**digits

    sx = np.ceil(sci_coeff * dig_scale) * sci_scale / dig_scale
    return sx


def signif_floor(x, digits: int = 3):
    """
    x = sci_coeff * 10**sci_exp
    """
    x = np.atleast_1d(x)
    idx = np.abs(x) < np.finfo(float).eps
    x[idx] = 3 * np.finfo(float).eps
    # scientific notation
    sci_exp = np.array([int(np.floor(np.log10(np.abs(x))))])
    sci_scale = 10.0**sci_exp
    sci_coeff = x / sci_scale

    # number of digits
    dig_scale = 10.0**digits

    sx = np.floor(sci_coeff * dig_scale) * sci_scale / dig_scale
    return sx



import numpy as np
import torch

from harlow.utils.transforms import (
    ChainTransform,
    Identity,
    Normalize,
    Standardize,
    TensorTransform,
)

N_features = 5
N_outputs = 3
N_points = 10
N_batch = 2

test_X = np.random.rand(N_batch, N_points, N_features) * 1000
test_y = np.random.rand(N_batch, N_points, N_outputs) * 1000


def test_identity_transform():
    transform = Identity()
    transformed_X = transform.forward(test_X)
    untransformed_X = transform.reverse(transformed_X)
    assert np.allclose(transformed_X, test_X)
    assert np.allclose(untransformed_X, test_X)


def test_standardize_transform():
    transform = Standardize()
    transformed_X = transform.forward(test_X)
    untransformed_X = transform.reverse(transformed_X)
    assert np.allclose(transformed_X.mean(axis=-2), 0.0)
    assert np.allclose(transformed_X.std(axis=-2), 1.0)
    assert np.allclose(untransformed_X.mean(axis=-2), test_X.mean(axis=-2))
    assert np.allclose(untransformed_X.std(axis=-2), test_X.std(axis=-2))


def test_normalize_transform():
    transform = Normalize()
    transformed_X = transform.forward(test_X)
    untransformed_X = transform.reverse(transformed_X)
    assert np.allclose(transformed_X.min(axis=-2), 0.0)
    assert np.allclose(transformed_X.max(axis=-2), 1.0)
    assert np.allclose(untransformed_X.min(axis=-2), test_X.min(axis=-2))
    assert np.allclose(untransformed_X.max(axis=-2), test_X.max(axis=-2))


def test_tensor_transform():
    transform = TensorTransform()
    transformed_X = transform.forward(test_X)
    untransformed_X = transform.reverse(transformed_X)
    assert torch.is_tensor(transformed_X)
    assert isinstance(untransformed_X, np.ndarray)


def test_chain_transform():
    transform = ChainTransform(
        Identity(), Normalize(), Standardize(), TensorTransform()
    )
    transformed_X = transform.forward(test_X)
    untransformed_X = transform.reverse(transformed_X)
    assert torch.is_tensor(transformed_X)
    assert isinstance(untransformed_X, np.ndarray)



import itertools
import math
import os
import pickle
import time
from copy import deepcopy
from timeit import default_timer as timer
from typing import Callable, Optional, Tuple

# import numba as nb
import numpy as np
from loguru import logger
from scipy.spatial.distance import cdist
from sklearn.metrics import mean_squared_error
from tensorboardX import SummaryWriter

from harlow.sampling.sampling_baseclass import Sampler
from harlow.surrogating.surrogate_model import Surrogate
from harlow.utils.helper_functions import evaluate, latin_hypercube_sampling, rmse
from harlow.utils.numba_utils import euclidean_distance, np_any, np_min

# from harlow.distance import pdist_full_matrix

# TODO
#  * improve logging
#  * pretty timedelta: https://gist.github.com/thatalextaylor/7408395

nopython = True
fastmath = True


# -----------------------------------------------------
# USER FACING API (class)
# -----------------------------------------------------
# TODO: is this class really needed or just an unnecessary complication? it has a
#  single method
class LolaVoronoi(Sampler):
    """Lola-Vornoi adaptive design strategy for global surrogate modelling.

    The algorithm is proposed and described in this paper:
    [1] Crombecq, Karel, et al. (2011) A novel hybrid sequential design strategy for global
    surrogate modeling of computer experiments. SIAM Journal on Scientific Computing 33.4
    (2011): 1948-1974.

    The implementation is influenced by:
    * gitlab.com/energyincities/besos/-/blob/master/besos/
    * https://github.com/FuhgJan/StateOfTheArtAdaptiveSampling/blob/master/src/adaptive_techniques/LOLA_function.m  # noqa E501

    """

    def __init__(
        self,
        target_function: Callable[[np.ndarray], np.ndarray],
        surrogate_model: Surrogate,
        domain_lower_bound: np.ndarray,
        domain_upper_bound: np.ndarray,
        fit_points_x: np.ndarray = None,
        fit_points_y: np.ndarray = None,
        test_points_x: np.ndarray = None,
        test_points_y: np.ndarray = None,
        evaluation_metric: Callable = None,
        run_name: str = "",
        save_dir: str = "",
    ):
        self.domain_lower_bound = domain_lower_bound
        self.domain_upper_bound = domain_upper_bound
        self.target_function = lambda x: target_function(x).reshape((-1, 1))
        self.surrogate_model = surrogate_model
        self.fit_points_x = fit_points_x
        self.fit_points_y = fit_points_y
        self.test_points_x = test_points_x
        self.test_points_y = test_points_y

        if evaluation_metric is None:
            self.metric = [rmse]
        else:
            self.metric = evaluation_metric
        self.run_name = run_name
        self.save_dir = save_dir
        # self.verbose = verbose

        # Internal storage for inspection
        self.step_x = []
        self.step_y = []
        self.step_score = []
        self.step_iter = []
        self.step_fit_time = []
        self.step_gen_time = []
        # Init writer for live web-based logging.
        self.writer = SummaryWriter(comment="-" + run_name)

    def sample(
        self,
        n_initial_point: int = None,
        n_iter: int = 20,
        n_new_point_per_iteration: int = 1,
        stopping_criterium: float = None,
        ignore_far_neighborhoods: Optional[bool] = True,
        ignore_old_neighborhoods: Optional[bool] = True,
    ):
        """TODO: allow for providing starting points"""
        # ..........................................
        # Initialize
        # ..........................................
        target_function = self.target_function
        domain_lower_bound = self.domain_lower_bound
        domain_upper_bound = self.domain_upper_bound
        n_dim = len(domain_lower_bound)

        if n_initial_point is None:
            n_initial_point = 5 * n_dim

        if stopping_criterium:
            n_iter = 1000

        if stopping_criterium and not self.metric:
            self.metric = lambda x, y: math.sqrt(mean_squared_error(x, y))

        # ..........................................
        # Initial sample of points
        # ..........................................
        gen_start_time = time.time()
        if self.fit_points_x is None:
            # latin hypercube sampling to get the initial sample of points
            points_x = latin_hypercube_sampling(
                n_sample=n_initial_point,
                domain_lower_bound=domain_lower_bound,
                domain_upper_bound=domain_upper_bound,
            )
            # evaluate the target function
            points_y = target_function(points_x)

            self.fit_points_x = points_x
            self.fit_points_y = points_y
        else:
            points_x = self.fit_points_x
            points_y = self.fit_points_y

        self.step_gen_time.append(time.time() - gen_start_time)

        # fit the surrogate model
        start_time = time.time()
        self.surrogate_model.fit(points_x, points_y.ravel())
        logger.info(
            f"Fitted the first surrogate model in {time.time() - start_time} sec."
        )

        # Additional class objects to help keep track of the sampling. `gen_time`
        # denotes the time to generate new points. `fit_time` is the time
        # to fit the surrogate.
        score = evaluate(
            self.metric, self.surrogate_model, self.test_points_x, self.test_points_y
        )
        self.step_x.append(points_x)
        self.step_y.append(points_y)
        self.step_score.append(score)
        self.step_iter.append(0)
        self.step_fit_time.append(time.time() - start_time)
        # ..........................................
        # Iterative improvement (adaptive stage)
        # ..........................................
        n_point_last_iter = 0
        for ii in range(n_iter):
            logger.info(
                f"Started adaptive iteration step: {ii + 1} (max steps:" f" {n_iter})."
            )

            start_time = time.time()
            new_points_x = best_new_points(
                points_x=points_x,
                points_y=points_y,
                domain_lower_bound=domain_lower_bound,
                domain_upper_bound=domain_upper_bound,
                n_point_last_iter=n_point_last_iter,
                n_new_point=n_new_point_per_iteration,
                ignore_far_neighborhoods=ignore_far_neighborhoods,
                ignore_old_neighborhoods=ignore_old_neighborhoods,
            )
            self.step_gen_time.append(time.time() - start_time)
            n_point_last_iter = points_x.shape[0]
            logger.info(
                f"Found the next best {n_new_point_per_iteration} point(s) in "
                f"{time.time() - start_time} sec."
            )

            # evaluate the target function
            new_points_y = target_function(new_points_x)

            # add the new points to the old ones
            points_x = np.vstack((points_x, new_points_x))
            points_y = np.vstack((points_y, new_points_y))

            # refit the surrogate
            start_time = time.time()
            self.surrogate_model.update(new_points_x, new_points_y.ravel())
            self.step_fit_time = time.time() - start_time
            logger.info(
                f"Fitted a new surrogate model in {time.time() - start_time} sec."
            )

            #
            self.fit_points_x = points_x
            self.fit_points_y = points_y
            score = evaluate(
                self.metric,
                self.surrogate_model,
                self.test_points_x,
                self.test_points_y,
            )
            self.step_x.append(points_x)
            self.step_y.append(points_y)
            self.step_score.append(score)
            self.step_iter.append(ii + 1)
            # Writer log
            if len(score) == 1:
                self.writer.add_scalar("RMSE", score[0], ii + 1)
            elif len(score) == 2:
                self.writer.add_scalar("RMSE", score[0], ii + 1)
                self.writer.add_scalar("RRSE", score[1], ii + 1)
            elif len(score) == 3:
                self.writer.add_scalar("RMSE", score[0], ii + 1)
                self.writer.add_scalar("RRSE", score[1], ii + 1)
                self.writer.add_scalar("MAE", score[2], ii + 1)
            self.writer.add_scalar("Gen time", self.step_gen_time[ii], ii + 1)
            self.writer.add_scalar("Fit time", self.step_fit_time, ii + 1)
            logger.info(f"Evaluation metric score on provided testset: {score}")
            self.score = score[0]
            self.iterations = ii
            # Save model every 5 iterations
            if self.iterations % 5 == 0:
                self.save_model()
            if stopping_criterium:
                if self.score <= stopping_criterium:
                    logger.info(f"Algorithm converged in {ii} iterations")
                    # Save model if converged
                    self.save_model()
                    break
            if n_iter:
                if ii > n_iter:
                    logger.info(f"Max number of iterations ({n_iter}) " f"reached.")
                    self.save_model()
                    break
        self.writer.close()
        return self.fit_points_x, self.fit_points_y

    def save_model(self):
        save_name = self.run_name + "_{}_iters.pkl".format(self.iterations)
        save_path = os.path.join(self.save_dir, save_name)

        with open(save_path, "wb") as file:
            pickle.dump(self.surrogate_model, file)

    def result_as_dict(self):
        pass


# -----------------------------------------------------
# SUPPORTING FUNCTIONS
# -----------------------------------------------------
def best_new_points(
    points_x: np.ndarray,
    points_y: np.ndarray,
    domain_lower_bound: np.ndarray,
    domain_upper_bound: np.ndarray,
    n_point_last_iter: int,
    n_new_point: int = 1,
    ignore_far_neighborhoods: Optional[bool] = True,
    ignore_old_neighborhoods: Optional[bool] = True,
):
    # shape the input if not in the right shape
    n_dim = len(domain_lower_bound)
    points_x = points_x.reshape((-1, n_dim))
    points_y = points_y.reshape((-1, 1))

    # Find the best neighborhoods for each row of `points_x`
    (
        best_neighborhood_scores,
        best_neighborhood_idxs,
        all_neighborhood_scores,
        all_neighbor_point_idxs_combinations,
    ) = best_neighborhoods(
        points_x,
        n_point_last_iter,
        n_new_point,
        ignore_far_neighborhoods=ignore_far_neighborhoods,
        ignore_old_neighborhoods=ignore_old_neighborhoods,
    )

    # Find the `n_new_point_per_step` best next/new point(s)
    idx = all_neighbor_point_idxs_combinations[best_neighborhood_idxs]
    all_best_neighbor_points_x = points_x[idx, :]
    all_best_neighbor_points_y = points_y[idx, :]

    new_points_x = best_new_points_with_neighbors(
        reference_points_x=points_x,
        reference_points_y=points_y,
        all_neighbor_points_x=all_best_neighbor_points_x,
        all_neighbor_points_y=all_best_neighbor_points_y,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        n_next_point=n_new_point,
    )

    return new_points_x


def best_new_points_with_neighbors(
    reference_points_x: np.ndarray,
    reference_points_y: np.ndarray,
    all_neighbor_points_x: np.ndarray,
    all_neighbor_points_y: np.ndarray,
    domain_lower_bound: np.ndarray,
    domain_upper_bound: np.ndarray,
    n_next_point: int,
):
    """Find the new/next reference point(s) where the target function will be evaluated.

    TODO: reuse the random point used here for other calls of this function (
        `best_next_points`)
    """
    n_dim = reference_points_x.shape[1]

    nonlinearity_measures = lola_score(
        all_neighbor_points_x=all_neighbor_points_x,
        all_neighbor_points_y=all_neighbor_points_y,
        reference_points_x=reference_points_x,
        reference_points_y=reference_points_y,
    )
    (
        relative_volumes,
        random_points,
        distance_mx,
        closest_indicator_mx,
    ) = voronoi_volume_estimate(
        points=reference_points_x,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
    )
    hybrid_score = lola_voronoi_score(nonlinearity_measures, relative_volumes)

    # TODO: check np.partition as an alternative
    idxs_new_neighbor = np.argsort(-hybrid_score)[:n_next_point]

    new_reference_points_x = np.empty((n_next_point, n_dim))
    for ii, idx_new_neighbor in enumerate(idxs_new_neighbor):
        # all the distances to reference point whose neighborhood will get a new
        # point where the target function is evaluated
        distances_in_neighbor = distance_mx[:, idx_new_neighbor]

        # consider only those points that are within the voronoi cell of the
        # reference point
        idx_mask_in_neighbor = closest_indicator_mx[:, idx_new_neighbor]

        # to avoid selecting points that are not in the neighborhood
        distances_in_neighbor[~idx_mask_in_neighbor] = -1

        # largest distance within the same voronoi cell
        idx_max_distance = np.argmax(distances_in_neighbor)
        new_reference_point_x = random_points[idx_max_distance]
        new_reference_points_x[ii, :] = new_reference_point_x

    return new_reference_points_x


def best_neighborhoods(
    points_x: np.ndarray,
    n_point_last_iter: np.ndarray,
    n_new_point_per_iter,
    ignore_far_neighborhoods: Optional[bool] = True,
    ignore_old_neighborhoods: Optional[bool] = True,
):
    """Find the best neighborhood for each row of `points_x`. This function is expected
    to be used with the initial sample of `points_x`, when no best neighborhoods yet
    available from preceding iteration steps. This function exists to separate numba
    compatible code from non-compatible one."""

    # n-choose-k, all possible neighbor combinations, the elements of the matrix are
    # `points_x` indices
    n_point, n_dim = points_x.shape
    n_neighbor = 2 * n_dim

    # All neighborhoods
    all_neighbor_point_idxs_combinations = np.array(
        list(itertools.combinations(np.arange(n_point), n_neighbor))
    )

    # Neighborhoods considering one less point
    neighbor_point_combinations = np.array(
        list(itertools.combinations(np.arange(n_point - 1), n_neighbor))
    )

    (
        best_neighborhood_scores,
        best_neighborhood_idxs,
        all_neighborhood_scores,
    ) = best_neighborhoods_numba(
        points_x,
        neighbor_point_combinations,
        n_point_last_iter,
        n_new_point_per_iter,
        ignore_far_neighborhoods=ignore_far_neighborhoods,
        ignore_old_neighborhoods=ignore_old_neighborhoods,
    )

    return (
        best_neighborhood_scores,
        best_neighborhood_idxs,
        all_neighborhood_scores,
        all_neighbor_point_idxs_combinations,
    )


def best_neighborhoods_numba(
    points_x: np.ndarray,
    neighbor_point_combinations: np.ndarray,
    n_point_last_iter: int,
    n_new_point_per_iter,
    ignore_far_neighborhoods: Optional[bool] = True,
    ignore_old_neighborhoods: Optional[bool] = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Find the best neighborhood for each row of `points_x`. This function is expected
    to be used with the initial sample of `points_x`, when no best neighborhoods yet
    available from preceding iteration steps.

    Args:
        points_x: n_point x n_dim.
        neighbor_point_combinations:
        n_point_last_iter:
        ignore_far_neighborhoods:
        ignore_old_neighborhoods:
    Returns:

    TODO:
     * It may be possible to refactor this part. Instead of creating the
     array of all valid combinations for N points, create the array for
     N-1 points. At each iteration of this loop we ignore the reference
     point, meaning that every time we can use the same combination of
     N-1 points, but with different indices. The indices in the array of
     combinations can be considered as 'labels'. Instead of excluding all
     combinations with the reference point at each iteration (which is costly),
     we can simply change the mapping between 'labels' and points.
     * The neighborhood score for a given reference point can not decrease
      between iterations. This allows us to only store a single best score
      for each point, instead of a n_neighborhood x n_point array of scores.

    """

    # TODO: It should be possible to perform the calculation of
    # all neighbourhoods after discarding the points that are
    # too far away. This will reduce storage requirements
    # and the computational cost of obtaining all combinations.
    n_point, n_dim = points_x.shape
    n_neighborhood = neighbor_point_combinations.shape[0]
    idx_valid_neighborhoods = np.arange(n_neighborhood)

    # Number of new points
    n_new_points = points_x.shape[0] - n_point_last_iter

    # this matrix contains neighborhoods that contain the reference point as well, these
    # will be assigned -1 and the rest will get a neighborhood score
    # each column belong to one point, same order as points_x (reference_point)
    all_neighborhood_scores = -np.ones((n_neighborhood, n_point))

    # TODO: both loops are completely independent hence parallelizable
    # for ii in nb.prange(n_point):
    # This loop is executed for each point in the domain
    t_neighbour = []
    t_test = []
    t_ignore_far = []
    t1_full = timer()

    # Array of labels
    arr_labels = np.arange(n_point - 1)

    # Array of new point indices
    arr_new_point_no = np.arange(n_point_last_iter, n_point_last_iter + n_new_points)

    for ii in range(n_point):

        reference_point_x = np.expand_dims(points_x[ii], 0)

        # # consider only the valid neighborhoods: the ones that do not contain
        # # `reference_point_x`
        # idx_valid_neighborhoods = np.where(
        #     np_all(all_neighbor_point_idxs_combinations != ii, axis=1)
        # )[0]
        #
        # arr_valid_neighbourhoods = all_neighbor_point_idxs_combinations[
        #     idx_valid_neighborhoods
        # ]

        # Array containing all point indices except for reference point
        arr_points = np.arange(n_point - 1)
        arr_points[ii:] = arr_points[ii:] + 1

        # Coordinates of all points except reference point
        arr_x = points_x[arr_points]

        # Make a copy of the array of neighborhoods
        arr_valid_neighbourhoods = deepcopy(neighbor_point_combinations)

        t1_test = timer()
        if ignore_old_neighborhoods:

            # Get array of labels of new points
            arr_new_point_labels = get_point_labels(
                arr_new_point_no, arr_labels, arr_points
            )

            # Remove entries from valid neighborhoods that
            # do not contain a new point
            arr_valid_neighbourhoods_mask = np.repeat(
                False, len(neighbor_point_combinations)
            )

            # Iterate over the new point labels and find all neighborhoods
            # that contain a new point
            for _idx_pt, pt in enumerate(arr_new_point_labels):

                # Mask valid neighbourhoods (i.e. neighbourhoods that contain
                # the current new point). Terminate early if all elements
                # in mask array are true.
                arr_valid_neighbourhoods_mask = np.bitwise_or(
                    arr_valid_neighbourhoods_mask,
                    np_any(neighbor_point_combinations == pt, axis=1),
                )
                if np.all(arr_valid_neighbourhoods_mask):
                    break
            arr_valid_neighbourhoods = neighbor_point_combinations[
                arr_valid_neighbourhoods_mask
            ]
            idx_valid_neighborhoods = np.where(arr_valid_neighbourhoods_mask)[0]
        t2_test = timer()
        t_test.append(t2_test - t1_test)

        t1_ignore_far = timer()
        # consider only the new neighborhoods that are not too far away.
        # neighborhoods that contain a point farther away the median Euclidean distance
        # are dismissed.
        # section [5] of the paper
        if np.where(idx_valid_neighborhoods)[0].size:
            if ignore_far_neighborhoods:
                all_neighbor_points_x = arr_x[arr_valid_neighbourhoods, :]
                dists = np.linalg.norm(
                    all_neighbor_points_x - reference_point_x, axis=-1
                )
                med = np.median(dists)
                idx_valid_neighborhoods = idx_valid_neighborhoods[
                    np.all(dists <= med, axis=-1)
                ]
        t2_ignore_far = timer()
        t_ignore_far.append(t2_ignore_far - t1_ignore_far)

        # compute the neighbourhood_score (`ns`) for each neighborhood
        t1_n = timer()
        if np.where(idx_valid_neighborhoods)[0].size:
            for idx_valid_neighborhood in idx_valid_neighborhoods:
                neighbor_points_x = arr_x[
                    neighbor_point_combinations[idx_valid_neighborhood], :
                ]
                # this is a time consuming function
                ns = neighborhood_score(
                    neighbor_points_x=neighbor_points_x,
                    reference_point_x=reference_point_x,
                )[0]
                all_neighborhood_scores[idx_valid_neighborhood, ii] = ns
        t2_n = timer()
        t_neighbour.append(t2_n - t1_n)
    t2_full = timer()

    print(f"Time for neighbourhood scores = {np.sum(t_neighbour)}")
    print(f"Test time = {np.sum(t_test)}")
    print(f"Ignore time = {np.sum(t_ignore_far)}")
    print(f"Full time = {t2_full - t1_full}")
    # best neighborhoods, the order is the same as in `points_x`
    # best_neighborhood_idxs = np_argmax(all_neighborhood_scores, axis=0)
    best_neighborhood_idxs = np.argmax(all_neighborhood_scores, axis=0)
    flat_idxs = n_point * best_neighborhood_idxs + np.arange(n_point)
    # best_neighborhood_scores = all_neighborhood_scores.ravel()[
    #     flat_idxs.astype(nb.int_)
    # ]
    best_neighborhood_scores = all_neighborhood_scores.ravel()[flat_idxs.astype(np.int)]
    return (
        best_neighborhood_scores,
        best_neighborhood_idxs,
        all_neighborhood_scores,
    )


def get_point_labels(points, label_array, point_array):
    """
    Get the labels of a set of points given two arrays that
    define a mapping between labels and points.

    WARNING: This function will not work correctly if there are
    elements in `points` that do not exist in `point_array`.
    """

    return label_array[np.nonzero(np.isin(point_array, points))]


def lola_voronoi_score(
    nonlinearity_measures: np.ndarray, relative_volumes: np.ndarray
) -> np.ndarray:
    """Eq.(5.1) of [1]."""
    return relative_volumes + nonlinearity_measures / np.sum(nonlinearity_measures)


def neighborhood_score(
    neighbor_points_x: np.ndarray, reference_point_x: np.ndarray
) -> Tuple[float, float]:
    """

    Args:
        neighbor_points_x:
            n_point x n_dim.
        reference_point_x:
            1 x n_dim

    Returns:
        Neighborhood score.
    """
    # shapes are not checked!
    # shape the input if not in the right shape, TODO: is this really needed?
    reference_point_x = reference_point_x.reshape((1, -1))
    n_dim = reference_point_x.shape[1]
    # neighbor_points_x = neighbor_points_x.reshape((-1, n_dim))

    # cohesion, Eq(4.3) of [1]
    cohesion = np.mean(
        np.sqrt(np.sum((neighbor_points_x - reference_point_x) ** 2, axis=1))
    )

    # cross-polytope ratio
    if n_dim == 1:
        pr1 = neighbor_points_x[0, 0]
        pr2 = neighbor_points_x[1, 0]
        # Eq(4.6) of [1]
        cross_polytope_ratio = 1 - np.abs(pr1 + pr2) / (
            np.abs(pr1) + np.abs(pr2) + np.abs(pr1 - pr2)
        )
    else:
        # Eq(4.4) of [1]
        # smallest neighbor distance for each neighbor
        # neighbor_distances = squareform(pdist(neighbor_points_x, metric="euclidean"))
        # np.fill_diagonal(neighbor_distances, np.inf)
        # min_neighbor_distances = np.min(neighbor_distances, axis=1)
        #  neighbor_distances = pdist_full_matrix(neighbor_points_x)

        # This should be an equivalent alternative to squareform + pdist
        # that also works with numba. Based on this stackoverflow answer:
        # https://stackoverflow.com/questions/60839615/
        # efficiently-calculating-a-euclidean-dist-matrix-in-numpy
        # TODO: Check
        # TODO: would be good to avoid np.max
        # neighbor_distances = squareform(pdist(neighbor_points_x, metric="euclidean"))

        neighbor_distances = euclidean_distance(neighbor_points_x)
        np.fill_diagonal(neighbor_distances, np.max(neighbor_distances))
        # min_neighbor_distances = np_min(neighbor_distances, axis=1)
        min_neighbor_distances = np_min(neighbor_distances, axis=1)
        adhesion = np.mean(min_neighbor_distances)
        # Eq(4.5) of [1]
        cross_polytope_ratio = adhesion / (np.sqrt(2) * cohesion)

    # Eq(4.7) of [1]
    return cross_polytope_ratio / cohesion, cross_polytope_ratio


def lola_score(
    all_neighbor_points_x: np.ndarray,
    all_neighbor_points_y: np.ndarray,
    reference_points_x: np.ndarray,
    reference_points_y: np.ndarray,
) -> np.ndarray:
    """Non-linearity measure for each point and its neighbor. Measures how much a
    neighborhood deviates from a hyperplane."""
    n_reference_point = len(reference_points_y)
    es = np.empty(n_reference_point)

    for (
        ii,
        (
            neighbor_points_x,
            neighbor_points_y,
            reference_point_x,
            reference_point_y,
        ),
    ) in enumerate(
        zip(
            all_neighbor_points_x,
            all_neighbor_points_y,
            reference_points_x,
            reference_points_y,
        )
    ):
        reference_point_gradient = gradient_estimate(
            reference_point_x=reference_point_x,
            reference_point_y=reference_point_y,
            neighbor_points_x=neighbor_points_x,
            neighbor_points_y=neighbor_points_y,
        )[0]
        es[ii] = nonlinearity_measure(
            reference_point_x=reference_point_x,
            reference_point_y=reference_point_y,
            reference_point_gradient=reference_point_gradient,
            neighbor_points_x=neighbor_points_x,
            neighbor_points_y=neighbor_points_y,
        )

    return es


def nonlinearity_measure(
    reference_point_x: np.ndarray,
    reference_point_y: float,
    reference_point_gradient: np.ndarray,
    neighbor_points_x: np.ndarray,
    neighbor_points_y: np.ndarray,
) -> float:
    # Eq.(4.9) of [1]
    e = np.sum(
        np.abs(
            neighbor_points_y
            - (
                reference_point_y
                + reference_point_gradient * (neighbor_points_x - reference_point_x)
            )
        )
    )
    return float(e)


def voronoi_volume_estimate(
    points: np.ndarray,
    domain_lower_bound: np.ndarray,
    domain_upper_bound: np.ndarray,
    n_simulation: int = None,
    random_points: np.ndarray = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Estimate the relative volume of the Voronoi tessellation of `points` bounded by
    `domain_lower_bound` and `domain_upper_bound`.

    The algorithm is described in section 3 of [1].

    Args:
        points:
            n_point x n_dim.
        domain_lower_bound:
            n_dim.
        domain_upper_bound:
            n_dim
        n_simulation:
            Number of random points used to estimate the relative volumes. If
            `random_points` is provided then this argument is ignored.
        random_points:

    Returns:
        Relative volumes in the same order as `points`.
    """
    # dimensions are not checked
    if n_simulation is None:
        n_simulation = 1000 * points.shape[0]

    if random_points is None:
        n_dim = len(domain_lower_bound)
        random_points = domain_lower_bound + np.random.rand(n_simulation, n_dim) * (
            domain_upper_bound - domain_lower_bound
        )

    # all relevant distances, n_simulation x n_point
    distance_mx = cdist(random_points, points, metric="euclidean")

    # index of the closest `point` to each `random_point`
    col_idx_min = np.argmin(distance_mx, axis=1)
    row_idx_min = np.arange(distance_mx.shape[0])

    # indicator matrix to count the number of `random_points` closest to each `points`
    closest_indicator_mx = np.zeros(distance_mx.shape, dtype=bool)
    # place a one (True) in the indicator matrix if the element (distance) is a closest
    # distance
    closest_indicator_mx[row_idx_min, col_idx_min] = True

    # count the closest `random_points` for each `points`
    closest_counts = np.sum(closest_indicator_mx, axis=0)

    # relative volume estimate
    relative_volumes = closest_counts / n_simulation

    return relative_volumes, random_points, distance_mx, closest_indicator_mx


def gradient_estimate(
    reference_point_x: np.ndarray,
    reference_point_y: float,
    neighbor_points_x: np.ndarray,
    neighbor_points_y: np.ndarray,
) -> np.ndarray:
    """
    Estimate the gradient at `reference_point` by fitting a hyperplane to
    `neighbor_points` in a least-square sense. A hyperplane that goes exactly
    through the `reference_point`.

    We think that the there is a mistake in Eq.(4.8) of [1], the right hand side
    should be `f(p_neighbor) - f(p_reference)`, or on the left hand side the
    `-p_reference` should be dropped if the formulation is in line with this:
    "Without loss of generality, we assume that pr lies in the origin."
    section 4.2.1 of [1].

    Args:
        reference_point_x:
            1 x n_dim.
        reference_point_y:
        neighbor_points_x:
            n_neighbor x n_dim.
        neighbor_points_y:
            n_neighbor x 1.

    Returns:
        Gradient estimate, 1 x n_dim.
    """
    # shape the input if not in the right shape, TODO: is this really needed?
    reference_point_x = reference_point_x.reshape((1, -1))
    n_dim = reference_point_x.shape[1]
    neighbor_points_x = neighbor_points_x.reshape((-1, n_dim))
    neighbor_points_y = neighbor_points_y.reshape((-1, 1))

    # to ensure that we hyperplane goes through `reference_point`
    neighbor_points_x_diff = neighbor_points_x - reference_point_x
    neighbor_points_y_diff = neighbor_points_y - reference_point_y

    # least-square fit of the hyperplane, the gradient is the hyperplane coefficients
    gradient = np.linalg.lstsq(
        neighbor_points_x_diff, neighbor_points_y_diff, rcond=None
    )[0].reshape((1, n_dim))

    return gradient



"""NOTE: this file is not yet harmonized with the general changes in the packages so
it is excluded from the testing (see the file name)."""
import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import r2_score
from sklearn.preprocessing import MinMaxScaler

from harlow.sampling import LolaVoronoi, ProbabilisticSampler
from harlow.surrogating import GaussianProcessTFP, VanillaGaussianProcess
from harlow.utils.test_functions import bohachevsky_2D, forrester_1d, shekel
from harlow.visualization.plotting import add_samples_to_plot, plot_function_custom


def test_2D():
    domain = np.array([[-100.0, 100.0], [-100.0, 100.0]])
    n_points = 40
    n_iters = 10
    n_per_iters = 4
    show_plot = False

    X1 = np.random.uniform(domain[0, 0], domain[0, 1], n_points)
    X2 = np.random.uniform(domain[1, 0], domain[1, 1], n_points)
    X = np.stack([X1, X2], -1)

    indices = np.random.permutation(X1.shape[0])
    train_idx, test_idx = (
        indices[: round(len(indices) * 0.5)],
        indices[round(len(indices) * 0.5) :],
    )
    train_X = X[train_idx, :]
    test_X = X[test_idx, :]
    train_y = bohachevsky_2D(train_X)
    test_y = bohachevsky_2D(test_X)

    gp = GaussianProcessTFP()
    gp.fit(train_X, train_y)
    gp_copy = GaussianProcessTFP()
    gp_copy.fit(train_X, train_y)

    p = gp.predict(test_X)
    print(f"test R2: {r2_score(test_y, p)}")
    print(f"test2 R2: {r2_score(test_y, gp_copy.predict(test_X))}")

    plot = plot_function_custom(
        bohachevsky_2D,
        train_X,
        y_vec=gp.predict(train_X),
        plot_sample_locations=True,
        show=show_plot,
    )

    lv = LolaVoronoi(
        gp,
        train_X,
        train_y,
        test_X,
        test_y,
        domain,
        bohachevsky_2D,
        n_init=1,
        n_iteration=n_iters,
        n_new_point_per_iteration=n_per_iters,
    )
    lv.run_sequential_design()

    plt.plot(np.arange(0, n_iters + 1), lv.score)

    add_samples_to_plot(
        plot,
        lv.train_X[-n_iters * n_per_iters :],
        bohachevsky_2D(lv.train_X[-n_iters * n_per_iters :]),
        "g",
    )

    plot_function_custom(
        bohachevsky_2D,
        lv.train_X,
        lv.surrogate_model.predict(lv.train_X),
        plot_sample_locations=True,
        show=show_plot,
    )

    random_scores = [r2_score(test_y, p)]
    for _ in range(n_iters):
        X1_random_test = np.random.uniform(domain[0, 0], domain[0, 1], n_per_iters)
        X2_random_test = np.random.uniform(domain[1, 0], domain[1, 1], n_per_iters)
        X_random_test = np.stack([X1_random_test, X2_random_test], -1)
        train_X = np.concatenate([train_X, X_random_test])
        train_y = np.concatenate([train_y, bohachevsky_2D(X_random_test)])

        gp_copy.update(X_random_test, bohachevsky_2D(X_random_test))
        rand_y = gp_copy.predict(test_X)
        random_scores.append(r2_score(test_y, rand_y))

    plt.plot(np.arange(0, n_iters + 1), random_scores)

    plot_function_custom(
        bohachevsky_2D, train_X, train_y, plot_sample_locations=True, show=show_plot
    )


def test_1D():
    domain = np.array([0.0, 1.0])
    n_points = 10
    n_iters = 5
    n_per_iters = 3

    X_range = np.linspace(0, 1, 1000)
    y_range = forrester_1d(X_range)
    X = np.random.uniform(domain[0], domain[1], n_points)

    indices = np.random.permutation(X.shape[0])
    train_idx, test_idx = (
        indices[: round(len(indices) * 0.5)],
        indices[round(len(indices) * 0.5) :],
    )
    train_X = np.sort(X[train_idx])
    test_X = np.sort(X[test_idx])
    train_y = forrester_1d(train_X)
    test_y = forrester_1d(test_X)

    gp = GaussianProcessTFP()
    gp.fit(train_X.reshape(-1, 1), train_y)

    p = gp.predict(test_X.reshape(-1, 1))
    print(f"test R2: {r2_score(test_y, p)}")

    plot = plot_function_custom(
        forrester_1d,
        train_X.reshape(-1, 1),
        y_vec=gp.predict(train_X.reshape(-1, 1)),
        plot_sample_locations=True,
        show=False,
    )
    plot.plot(X_range, y_range, "r")

    lv = LolaVoronoi(
        gp,
        train_X.reshape(-1, 1),
        train_y,
        test_X.reshape(-1, 1),
        test_y,
        [domain],
        forrester_1d,
        n_iteration=n_iters,
        n_new_point_per_iteration=n_per_iters,
        evaluation_metric="rmse",
    )
    lv.run_sequential_design()

    add_samples_to_plot(
        plot,
        lv.train_X[-n_iters * n_per_iters :],
        forrester_1d(lv.train_X[-n_iters * n_per_iters :]),
        "g",
    )


def test_tfdGP():
    num_training_points = 100
    index_points_ = np.random.uniform(-1.0, 1.0, (num_training_points, 1))
    index_points_ = index_points_.astype(np.float64)
    observations_ = (np.sin(3 * np.pi * index_points_[..., 0])) + np.random.normal(
        loc=0, scale=np.sqrt(0.1), size=(num_training_points)
    )

    gp = GaussianProcessTFP()
    gp.fit(index_points_, observations_)

    predictive_index_points_ = np.linspace(-1.2, 1.2, 200, dtype=np.float64)
    predictive_index_points_ = predictive_index_points_[..., np.newaxis]
    mean, samples = gp.predict(predictive_index_points_, return_samples=True)

    plt.figure(figsize=(12, 4))
    plt.plot(
        predictive_index_points_,
        (np.sin(3 * np.pi * predictive_index_points_[..., 0])),
        label="True fn",
    )
    plt.scatter(index_points_[:, 0], observations_, label="Observations")
    for i in range(50):
        plt.plot(
            predictive_index_points_,
            samples[i, :],
            c="r",
            alpha=0.1,
            label="Posterior Sample" if i == 0 else None,
        )
    leg = plt.legend(loc="upper right")
    for lh in leg.legendHandles:
        lh.set_alpha(1)
    plt.xlabel(r"Index points ($\mathbb{R}^1$)")
    plt.ylabel("Observation space")
    plt.show()


def test_shekel():
    domain = np.array([[0.0, 10.0], [0.0, 10.0], [0.0, 10.0], [0.0, 10.0]])
    n_points = 40

    X1 = np.random.uniform(domain[0, 0], domain[0, 1], n_points)
    X2 = np.random.uniform(domain[1, 0], domain[1, 1], n_points)
    X3 = np.random.uniform(domain[2, 0], domain[2, 1], n_points)
    X4 = np.random.uniform(domain[3, 0], domain[3, 1], n_points)

    X1 = np.append(X1, 4.0)
    X2 = np.append(X2, 4.0)
    X3 = np.append(X3, 4.0)
    X4 = np.append(X4, 4.0)

    X = np.stack([X1, X2, X3, X4], -1)

    y = shekel(X)

    print(y)


def visual_test_probSampling_1D():
    domain = np.array([0.0, 1.0])
    n_points = 10
    X_range = np.linspace(0, 1, 1000)
    y_range = forrester_1d(X_range)
    X = np.random.uniform(domain[0], domain[1], n_points)

    indices = np.random.permutation(X.shape[0])
    train_idx, test_idx = (
        indices[: round(len(indices) * 0.5)],
        indices[round(len(indices) * 0.5) :],
    )
    train_X = np.sort(X[train_idx])
    test_X = np.sort(X[test_idx])
    train_y = forrester_1d(train_X).reshape((-1, 1))
    test_y = forrester_1d(test_X)

    scaler = MinMaxScaler()
    train_X = scaler.fit_transform(train_X.reshape(-1, 1))
    test_X = scaler.transform(test_X.reshape(-1, 1))

    gp = VanillaGaussianProcess()
    gp.fit(train_X, train_y)

    p = gp.predict(test_X)
    print(f"test R2: {r2_score(test_y, p)}")

    plot = plot_function_custom(
        forrester_1d,
        train_X.reshape(-1, 1),
        y_vec=gp.predict(train_X.reshape(-1, 1)),
        plot_sample_locations=True,
        show=False,
    )
    plot.plot(X_range, y_range, "r")

    gpr = VanillaGaussianProcess()
    lv = ProbabilisticSampler(
        target_function=forrester_1d,
        surrogate_model=gpr,
        domain_lower_bound=np.array([0.0]),
        domain_upper_bound=np.array([1.0]),
        fit_points_x=train_X.reshape(-1, 1),
        fit_points_y=train_y,
        test_points_x=test_X.reshape(-1, 1),
        test_points_y=test_y,
    )

    points_x, points_y = lv.sample()

    add_samples_to_plot(
        plot,
        points_x[0 : len(train_X)],
        forrester_1d(points_x[0 : len(train_X)]),
        "r",
    )
    add_samples_to_plot(
        plot,
        points_x[-(lv.iterations * 1) :],
        forrester_1d(points_x[-(lv.iterations * 1) :]),
        "g",
    )

    plt.show()



"""Visualization functions."""
from typing import Callable

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm


def get_cmap(n: int, name="hsv"):
    return plt.cm.get_cmap(name, n)


def plot_function(
    f: Callable,
    domain: np.ndarray,
    grid_size: int = 1000,
    plot_name: int = None,
    show: bool = True,
    save: bool = False,
):
    if domain.shape[0] > 2:
        raise NotImplementedError

    x0_range = domain[0]
    x1_range = domain[1]

    x1_vec = np.linspace(x0_range[0], x0_range[1], grid_size)
    x2_vec = np.linspace(x1_range[0], x1_range[1], grid_size)

    x1_mx, x2_mx = np.meshgrid(x1_vec, x2_vec)
    z_mx = f(
        np.hstack(
            (
                x1_mx.reshape(grid_size * grid_size, 1),
                x2_mx.reshape(grid_size * grid_size, 1),
            )
        )
    )
    fig = plt.figure()
    ax = plt.axes(projection="3d")
    surface = ax.plot_trisurf(
        x1_mx[:, 0],
        x2_mx[:, 1],
        z_mx.reshape(grid_size, grid_size),
        cmap=cm.coolwarm,
        label=f.__name__ if plot_name is None else plot_name,
    )
    fig.colorbar(surface)

    if show:
        plt.show()
    if save:
        plt.savefig(f"{f.__name__ if plot_name is None else plot_name}.png")
    return plt


def plot_function_custom(
    f: Callable,
    x_mx: np.ndarray,
    y_vec: np.ndarray = None,
    plot_sample_locations: bool = False,
    plot_name: str = None,
    show: bool = True,
    save: bool = False,
    color: str = "b",
):

    if y_vec is None:
        y_vec = f(x_mx)

    fig = plt.figure()

    if len(x_mx.squeeze().shape) == 1:
        ax = fig.gca()
        ax.plot(x_mx, y_vec, label=f.__name__ if plot_name is None else plot_name)
    else:
        ax = plt.axes(projection="3d")
        surface = ax.plot_trisurf(
            x_mx[:, 0],
            x_mx[:, 1],
            y_vec,
            cmap=cm.coolwarm,
            label=f.__name__ if plot_name is None else plot_name,
        )
        fig.colorbar(surface)

    if plot_sample_locations:
        if len(x_mx.shape) == 1:
            ax.scatter(x_mx, y_vec, c=color, alpha=0.6)
        elif x_mx.shape[1] > 1:
            ax.scatter(x_mx[:, 0], x_mx[:, 1], y_vec, c=color, alpha=0.6)

    if show:
        plt.show()
    if save:
        plt.savefig(f"{f.__name__ if plot_name is None else plot_name}.png")
    return plt


def add_samples_to_plot(plot, x_mx: np.ndarray, y_vec: np.ndarray, color: str):
    ax = plot.gca()

    if x_mx.shape[1] == 1:
        ax.scatter(x_mx, y_vec, c=color)
    elif x_mx.shape[1] > 1:
        ax.scatter(x_mx[:, 0], x_mx[:, 1], y_vec, c=color)

    return plt



from harlow.sampling.bayesian_optimization import (  # noqa F401
    NegativeIntegratedPosteriorVarianceSampler,
)
from harlow.sampling.fuzzy_lolavoronoi import FuzzyLolaVoronoi  # noqa F401
from harlow.sampling.lola_voronoi import LolaVoronoi  # noqa F401
from harlow.sampling.probabilistic_sampling import ProbabilisticSampler  # noqa F401
from harlow.sampling.random_sampling import LatinHypercube  # noqa F401
from harlow.sampling.sampling_baseclass import Sampler  # noqa F401



import numpy as np

from harlow.utils.test_functions import ackley_nd


def test_ackley_nd():
    # test up to this dimension (a few is enough to check the generalization)
    n_max_dim = 6
    a = 20
    b = 0.2
    c = 2 * np.pi

    # the global minimum is known for all versions (dimensions)
    for ii in range(n_max_dim):
        x_mx = np.zeros(ii + 1)
        y_expected = 0.0
        y = ackley_nd(x_mx=x_mx, a=a, b=b, c=c)

        np.testing.assert_almost_equal(y_expected, y)

    # test a few points against this Matlab implementation:
    # https://www.sfu.ca/~ssurjano/Code/ackleym.html
    x_mx = np.array([[1, 2], [3, -4]])
    y_expected = np.array([5.422131717799510, 10.138626172095204])
    y = ackley_nd(x_mx=x_mx, a=a, b=b, c=c)
    np.testing.assert_almost_equal(y_expected, y)

    # points that have the same value irrespective of the dimension, against the same
    # Matlab implementation
    y_expected = np.array([3.625384938440363, 6.593599079287213, 11.013420717655567])
    for ii in range(n_max_dim):
        x_mx = np.tile(np.array([[1], [-2], [4]]), reps=(1, ii + 1))

        y = ackley_nd(x_mx=x_mx, a=a, b=b, c=c)

        np.testing.assert_almost_equal(y_expected, y)



"""
Comparison of different methods for reducing the optimzal neighbourhood
calculation time. The different methods are compared for a given random
seed to determine:

* The surrogating wall clock time.
* The loss of efficiency caused by the heuristic, if any.
* That the optimizations are implemented correctly.
"""

import math

import numpy as np
import torch
from sklearn.metrics import mean_squared_error

from harlow.model.model_twin_girder_betti import IJssel_bridge_model  # noqa:
from harlow.sampling.cv_voronoi import CVVoronoi
from harlow.sampling.fuzzy_lolavoronoi import FuzzyLolaVoronoi
from harlow.surrogating.surrogate_model import (
    ModelListGaussianProcess,
    VanillaGaussianProcess,
)
from harlow.utils.helper_functions import latin_hypercube_sampling

# I201
from harlow.utils.test_functions import peaks_2d_multivariate

domains_lower_bound = np.array([-8, -8])
domains_upper_bound = np.array([8, 8])
n_initial_point = 15
n_new_points_per_iteration = 1
rmse_criterium = 0.05
np.random.seed(123)
n_iter_sampling = 500
n_iter_runs = 100


def get_param_idx(params_dict):
    return {key: idx_key for idx_key, key in enumerate(params_dict)}


# ====================================================================
# INITIALIZE MODEL
# ====================================================================
# Set model parameters
meas_type = "fugro"
E = 210e6
max_elem_length = 2.0 * 1e3

# Sensor names and positions
sensor_names = ["H1_S", "H2_S", "H3_S", "H4_S", "H5_S", "H7_S", "H8_S", "H9_S", "H10_S"]
sensor_positions = [
    20.42,
    34.82,
    47.700,
    61.970,
    68.600,
    96.800,
    113.9,
    123.900,
    147.500,
]

# All parameters. This is the order that parameters will be expcted in
# within the various functions in this scrípt (e.g. model response function)
params_all = ["Kr1", "Kr2", "Kr3", "Kr4", "Kv"]

N_tasks = len(sensor_positions)
N_features = len(params_all)

# Parameters that are shared between models
params_common = [
    "Kv",
]

# Parameters that are not shared by all models
params_model = {
    "H1_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H2_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H3_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H4_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H5_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H7_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H8_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H9_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H10_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
}

# Define FE models and append to list
models_dict = {
    idx_model: IJssel_bridge_model(
        sname, E, max_elem_length=max_elem_length, truck_load=meas_type
    )
    for idx_model, sname in enumerate(sensor_names)
}
node_xs = models_dict[0].node_xs

# # ====================================================================
# # DOMAIN BOUNDS
# # ====================================================================

# Prior for first support rotational stiffness Kr1
Kr_low = 4.0
Kr_high = 10.0

# Ground truth and prior for K-brace spring vertical stiffness
Kv_low = 0.0
Kv_high = 8.0

c = -0.1754

# Bounds of the time domain
t_low = np.min(node_xs)
t_high = np.max(node_xs)

params_priors = {
    "Kv": {"dist": "uniform", "low": Kv_low, "high": Kv_high},
    "Kr1": {"dist": "uniform", "low": Kr_low, "high": Kr_high},
    "Kr2": {"dist": "uniform", "low": Kr_low, "high": Kr_high},
    "Kr3": {"dist": "uniform", "low": Kr_low, "high": Kr_high},
    "Kr4": {"dist": "uniform", "low": Kr_low, "high": Kr_high},
}

# Create domain bounds
domain_lower_bound = np.array([params_priors[param]["low"] for param in params_all])
domain_upper_bound = np.array([params_priors[param]["high"] for param in params_all])

# Bounds as tensor
bounds = torch.tensor(np.vstack([domain_lower_bound, domain_upper_bound]))

# Indices of all params
param_idx = get_param_idx(params_all)

# Build list of parameter indices per model
list_params = [
    [param_idx[param] for param in params_model[key] + params_common]
    for key in params_model.keys()
]


# # ====================================================================
# # MODEL FUNCTION
# # ====================================================================
def response(X, pts):

    # Initialize
    X = np.atleast_2d(X)
    N_x = X.shape[0]
    N_y = len(pts)
    res = np.zeros((N_x, N_y))

    print(f"Evaluating response function at {N_x} points for {N_y} outputs")

    # Iterate over parameter array
    for idx_x, x in enumerate(X):
        Kr1 = x[param_idx["Kr1"]]
        Kr2 = x[param_idx["Kr2"]]
        Kr3 = x[param_idx["Kr3"]]
        Kr4 = x[param_idx["Kr4"]]
        Kv = x[param_idx["Kv"]]

        # Rotational stiffness input array
        arr_Kr = np.repeat(np.array([Kr1, Kr2, Kr3, Kr4]), 2)
        arr_Kr = np.append(arr_Kr, np.zeros(4))

        # Iterate over models and evaluate
        for idx_t, model_key in enumerate(models_dict.keys()):

            # Get model and prediction t
            t = pts[idx_t]
            model = models_dict[model_key]

            # Evaluate each model in list
            res[idx_x, idx_t] = np.interp(
                t,
                model.node_xs,
                model.il_stress_truckload(c, lane="left", Kr=10**arr_Kr, Kv=10**Kv),
            )
    return res


# Model function taking only X as input, to be passed to the sampler.
def func_model(X):
    return response(X, sensor_positions)


def rmse(x, y):
    return math.sqrt(mean_squared_error(x, y))


def create_test_set(min_domain, max_domain, n):
    test_X = latin_hypercube_sampling(domain_lower_bound, domain_upper_bound, n)
    test_y = func_model(test_X)

    return test_X, test_y


def run_2D_test():
    fuzzy_sampling_results = []
    test_X, test_y = create_test_set(domain_lower_bound, domain_upper_bound, 500)

    start_points_X = latin_hypercube_sampling(
        domain_lower_bound, domain_upper_bound, n_initial_point
    )
    start_points_y = func_model(start_points_X)

    fuzzy_sampling_results.append(
        test_2D_cvVoronoi_sampling(
            start_points_X,
            start_points_y,
            domain_lower_bound,
            domain_upper_bound,
            test_X,
            test_y,
            n_iter_sampling,
        )
    )


def run_2D_FLV_test():
    fuzzy_sampling_results = []
    test_X, test_y = create_test_set(domains_lower_bound, domains_upper_bound, 500)

    start_points_X = latin_hypercube_sampling(
        domains_lower_bound, domains_upper_bound, n_initial_point
    )
    start_points_y = peaks_2d_multivariate(start_points_X)

    fuzzy_sampling_results.append(
        test_2D_FLV_sampling(
            start_points_X,
            start_points_y,
            domains_lower_bound,
            domains_upper_bound,
            test_X,
            test_y,
            n_iter_sampling,
        )
    )


def test_2D_cvVoronoi_sampling(
    start_points_X,
    start_points_y,
    domain_lower_bound,
    domain_upper_bound,
    test_X,
    test_y,
    n_iter,
):

    surrogate_model = VanillaGaussianProcess

    # ............................
    # Surrogating
    # ............................
    lv = CVVoronoi(
        target_function=func_model,
        surrogate_model=surrogate_model,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        fit_points_x=start_points_X,
        fit_points_y=start_points_y,
        test_points_x=test_X,
        test_points_y=test_y,
        run_name=None,
    )
    # main_start = time.time()
    lv.sample(
        n_initial_points=n_initial_point,
        n_new_points_per_iteration=n_new_points_per_iteration,
        max_n_iterations=200,
    )

    return {
        "step_x": lv.step_x,
        "step_y": lv.step_y,
        "score": lv.step_score,
        "iteration": lv.step_iter,
        "gen_time": lv.step_gen_time,
        "fit_time": lv.step_fit_time,
    }


def test_2D_FLV_sampling(
    start_points_X,
    start_points_y,
    domain_lower_bound,
    domain_upper_bound,
    test_X,
    test_y,
    n_iter,
):

    start_points_X = np.tile(start_points_X, (2, 1, 1))
    test_X = np.tile(test_X, (2, 1, 1))
    surrogate_model = ModelListGaussianProcess(
        ["m1", "m2"], list_params=[[0, 1], [0, 1]]
    )

    # ............................
    # Surrogating
    # ............................
    lv = FuzzyLolaVoronoi(
        target_function=peaks_2d_multivariate,
        surrogate_model=surrogate_model,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        fit_points_x=start_points_X,
        fit_points_y=start_points_y,
        test_points_x=test_X,
        test_points_y=test_y,
        run_name=None,
    )
    # main_start = time.time()
    lv.sample(
        n_initial_points=n_initial_point,
        n_new_points_per_iteration=n_new_points_per_iteration,
        max_n_iterations=20,
    )

    return {
        "step_x": lv.step_x,
        "step_y": lv.step_y,
        "score": lv.step_score,
        "iteration": lv.step_iter,
        "gen_time": lv.step_gen_time,
        "fit_time": lv.step_fit_time,
    }


if __name__ == "__main__":
    run_2D_test()
    # run_2D_FLV_test()



import numba as nb
import numpy as np

nopython = True
fastmath = True
cache = True


@nb.jit(nopython=nopython, fastmath=fastmath, cache=cache)
def np_apply_along_axis(func1d, axis, arr):
    """Until `axis` is supported by `numba`.
    Source: https://github.com/numba/numba/issues/1269#issuecomment-472574352"""
    assert arr.ndim == 2
    assert axis in [0, 1]
    if axis == 0:
        result = np.empty(arr.shape[1])
        for ii in range(len(result)):
            result[ii] = func1d(arr[:, ii])
    else:
        result = np.empty(arr.shape[0])
        for ii in range(len(result)):
            result[ii] = func1d(arr[ii, :])
    return result


@nb.jit(nopython=nopython, fastmath=fastmath, cache=cache)
def np_min(array, axis):
    return np_apply_along_axis(np.min, axis, array)


@nb.jit(nopython=nopython, fastmath=fastmath, cache=cache)
def np_all(array, axis):
    return np_apply_along_axis(np.all, axis, array)


@nb.jit(nopython=nopython, fastmath=fastmath, cache=cache)
def np_argmax(array, axis):
    return np_apply_along_axis(np.argmax, axis, array).astype(nb.int_)


@nb.jit(nopython=nopython, fastmath=fastmath, cache=cache)
def np_any(array, axis):
    return np_apply_along_axis(np.any, axis, array).astype(nb.bool_)


@nb.jit(nopython=nopython, fastmath=fastmath, cache=cache)
def euclidean_distance(X):
    arr_1 = np.expand_dims(X, axis=1)
    arr_2 = np.expand_dims(X, axis=0)
    Nd = arr_1.shape[-1]
    Np = arr_1.shape[0]
    arr = np.zeros((Np, Np, Nd))
    for i in range(Nd):
        arr[:, :, i] = (arr_1[:, :, i] - arr_2[:, :, i]) ** 2
    X_dist = np.sum(arr, axis=-1)
    return np.sqrt(X_dist)



import time
from typing import Callable

import numpy as np
from loguru import logger

from harlow.sampling.sampling_baseclass import Sampler
from harlow.utils.helper_functions import evaluate, latin_hypercube_sampling
from harlow.utils.log_writer import write_scores, write_timer
from harlow.utils.metrics import rmse


class LatinHypercube(Sampler):
    def __init__(
        self,
        target_function,
        surrogate_model,
        domain_lower_bound: np.ndarray,
        domain_upper_bound: np.ndarray,
        fit_points_x: np.ndarray = None,
        fit_points_y: np.ndarray = None,
        test_points_x: np.ndarray = None,
        test_points_y: np.ndarray = None,
        evaluation_metric: Callable = rmse,
        logging_metrics: list = None,
        verbose: bool = False,
        run_name: str = None,
        save_dir: str = "",
    ):
        super(LatinHypercube, self).__init__(
            target_function,
            surrogate_model,
            domain_lower_bound,
            domain_upper_bound,
            fit_points_x,
            fit_points_y,
            test_points_x,
            test_points_y,
            evaluation_metric,
            logging_metrics,
            verbose,
            run_name,
            save_dir,
        )

    def sample(
        self,
        n_initial_points: int = None,
        max_n_iterations: int = 20,
        n_new_points_per_iteration: int = 1,
        stopping_criterium: float = None,
    ):

        points_x = self.fit_points_x
        points_y = self.fit_points_y
        surrogate_model = self.surrogate_model
        start_time = time.time()
        if self.fit_points_x:
            surrogate_model.fit(points_x, points_y)
            logger.info(
                f"Fitted a new surrogate model in {time.time() - start_time} sec."
            )

        # Initial sampling
        if n_initial_points:
            X_new = latin_hypercube_sampling(
                n_sample=n_initial_points,
                domain_lower_bound=self.domain_lower_bound,
                domain_upper_bound=self.domain_upper_bound,
            )
            self.step_gen_time.append(time.time() - start_time)
            y_new = self.target_function(X_new)

            if points_x:
                points_x = np.concatenate((points_x, X_new))
                points_y = np.concatenate((points_y, y_new))
            else:
                points_x = X_new
                points_y = y_new

            surrogate_model.fit(points_x, points_y)

        score = evaluate(
            self.logging_metrics,
            self.test_points_x,
            self.test_points_y,
        )
        self.score = score[self.evaluation_metric.__name__]
        self.step_x.append(points_x)
        self.step_y.append(points_y)
        self.step_score.append(score)
        self.step_iter.append(0)
        self.step_fit_time.append(time.time() - start_time)

        iteration = 0
        while (self.score > stopping_criterium) and (iteration < max_n_iterations):
            start_time = time.time()
            X_new = latin_hypercube_sampling(
                n_sample=n_new_points_per_iteration,
                domain_lower_bound=self.domain_lower_bound,
                domain_upper_bound=self.domain_upper_bound,
            )
            self.step_gen_time.append(time.time() - start_time)
            y_new = self.target_function(X_new)
            points_x = np.concatenate((points_x, X_new))
            points_y = np.concatenate((points_y, y_new))

            start_time = time.time()
            surrogate_model.fit(points_x, points_y)
            logger.info(f"Fitted a surrogate model in {time.time() - start_time} sec.")
            self.step_fit_time.append(time.time() - start_time)

            self.fit_points_x = points_x
            self.fit_points_y = points_y
            score = evaluate(
                self.logging_metrics,
                self.test_points_x,
                self.test_points_y,
            )
            self.score = score[self.evaluation_metric.__name__]
            logger.info(f"Score {score}")
            self.step_x.append(points_x)
            self.step_y.append(points_y)
            self.step_score.append(score[0])
            self.step_iter.append(iteration + 1)
            timing_dict = {
                "Gen time": self.step_gen_time[iteration + 1],
                "Fit time": self.step_fit_time[iteration + 1],
            }
            write_scores(self.writer, score, iteration + 1)
            write_timer(self.writer, timing_dict, iteration + 1)

            iteration += 1

            if iteration >= max_n_iterations:
                break

        logger.info(f"Algorithm converged in {iteration} iterations")
        logger.info(f"Algorithm converged with score {score}")
        self.writer.close()
        return self.fit_points_x, self.fit_points_y



import itertools
import os
import pickle
import time
from typing import Callable, Optional, Tuple

import numpy as np
from loguru import logger
from scipy.spatial.distance import cdist

from harlow.sampling.sampling_baseclass import Sampler
from harlow.surrogating.surrogate_model import Surrogate
from harlow.utils.helper_functions import evaluate, latin_hypercube_sampling
from harlow.utils.log_writer import write_scores, write_timer
from harlow.utils.metrics import rmse
from harlow.utils.numba_utils import euclidean_distance, np_all, np_any, np_min

# from harlow.distance import pdist_full_matrix

# TODO
#  * improve logging
#  * pretty timedelta: https://gist.github.com/thatalextaylor/7408395

nopython = True
fastmath = True


# -----------------------------------------------------
# USER FACING API (class)
# -----------------------------------------------------
# TODO: is this class really needed or just an unnecessary complication? it has a
#  single method
class LolaVoronoi(Sampler):
    """Lola-Vornoi adaptive design strategy for global surrogate modelling.

    The algorithm is proposed and described in this paper:
    [1] Crombecq, Karel, et al. (2011) A novel hybrid sequential design strategy for global
    surrogate modeling of computer experiments. SIAM Journal on Scientific Computing 33.4
    (2011): 1948-1974.

    The implementation is influenced by:
    * gitlab.com/energyincities/besos/-/blob/master/besos/
    * https://github.com/FuhgJan/StateOfTheArtAdaptiveSampling/blob/master/src/adaptive_techniques/LOLA_function.m  # noqa E501

    """

    def __init__(
        self,
        target_function: Callable[[np.ndarray], np.ndarray],
        surrogate_model: Surrogate,
        domain_lower_bound: np.ndarray,
        domain_upper_bound: np.ndarray,
        fit_points_x: np.ndarray = None,
        fit_points_y: np.ndarray = None,
        test_points_x: np.ndarray = None,
        test_points_y: np.ndarray = None,
        evaluation_metric: Callable = rmse,
        logging_metrics: list = None,
        verbose: bool = False,
        run_name: str = None,
        save_dir: str = "",
    ):

        super(LolaVoronoi, self).__init__(
            target_function,
            surrogate_model,
            domain_lower_bound,
            domain_upper_bound,
            fit_points_x,
            fit_points_y,
            test_points_x,
            test_points_y,
            evaluation_metric,
            logging_metrics,
            verbose,
            run_name,
            save_dir,
        )

    def sample(
        self,
        n_initial_points: int = None,
        max_n_iterations: int = 20,
        n_new_points_per_iteration: int = 1,
        stopping_criterium: float = None,
        ignore_far_neighborhoods: Optional[bool] = True,
        ignore_old_neighborhoods: Optional[bool] = True,
    ):
        # ..........................................
        # Initialize
        # ..........................................
        target_function = self.target_function
        domain_lower_bound = self.domain_lower_bound
        domain_upper_bound = self.domain_upper_bound
        n_dim = len(domain_lower_bound)

        if n_initial_points is None:
            n_initial_points = 5 * n_dim

        # ..........................................
        # Initial sample of points
        # ..........................................
        gen_start_time = time.time()
        if self.fit_points_x is None:
            # latin hypercube sampling to get the initial sample of points
            points_x = latin_hypercube_sampling(
                n_sample=n_initial_points,
                domain_lower_bound=domain_lower_bound,
                domain_upper_bound=domain_upper_bound,
            )
            # evaluate the target function
            points_y = target_function(points_x)

            self.fit_points_x = points_x
            self.fit_points_y = points_y
        else:
            points_x = self.fit_points_x
            points_y = self.fit_points_y

        self.step_gen_time.append(time.time() - gen_start_time)

        # fit the surrogate model
        start_time = time.time()
        self.surrogate_model.fit(points_x, points_y)
        logger.info(
            f"Fitted the first surrogate model in {time.time() - start_time} sec."
        )

        score = evaluate(
            self.logging_metrics,
            self.test_points_x,
            self.test_points_y,
        )

        self.step_score.append(score)
        self.step_x.append(points_x)
        self.step_y.append(points_y)
        self.step_iter.append(0)

        self.step_fit_time.append(time.time() - start_time)
        # ..........................................
        # Iterative improvement (adaptive stage)
        # ..........................................
        n_point_last_iter = 0
        for ii in range(max_n_iterations):
            logger.info(
                f"Started adaptive iteration step: {ii+1} (max steps:"
                f" {max_n_iterations})."
            )

            start_time = time.time()
            new_points_x = best_new_points(
                points_x=points_x,
                points_y=points_y,
                domain_lower_bound=domain_lower_bound,
                domain_upper_bound=domain_upper_bound,
                n_point_last_iter=n_point_last_iter,
                n_new_point=n_new_points_per_iteration,
                ignore_far_neighborhoods=ignore_far_neighborhoods,
                ignore_old_neighborhoods=ignore_old_neighborhoods,
            )
            self.step_gen_time.append(time.time() - start_time)
            n_point_last_iter = points_x.shape[0]
            logger.info(
                f"Found the next best {n_new_points_per_iteration} point(s) in "
                f"{time.time() - start_time} sec."
            )

            # evaluate the target function
            new_points_y = target_function(new_points_x)

            # add the new points to the old ones
            points_x = np.vstack((points_x, new_points_x))
            points_y = np.vstack((points_y, new_points_y))

            # refit the surrogate
            start_time = time.time()
            self.surrogate_model.update(new_points_x, new_points_y)
            self.step_fit_time.append(time.time() - start_time)
            logger.info(
                f"Fitted a new surrogate model in {time.time() - start_time} sec."
            )

            #
            self.fit_points_x = points_x
            self.fit_points_y = points_y

            score = evaluate(
                self.logging_metrics,
                self.test_points_x,
                self.test_points_y,
            )

            self.step_score.append(score)
            self.step_x.append(points_x)
            self.step_y.append(points_y)
            self.step_iter.append(ii + 1)
            timing_dict = {
                "Gen time": self.step_gen_time[ii + 1],
                "Fit time": self.step_fit_time[ii + 1],
            }
            write_scores(self.writer, score, ii + 1)
            write_timer(self.writer, timing_dict, ii + 1)

            logger.info(f"Evaluation metric score on provided testset: {score}")
            self.score = score[self.evaluation_metric.__name__]
            self.iterations = ii
            # Save model every 5 iterations
            if self.iterations % 5 == 0:
                self.save_model()

            if len(self.score) > 0:
                if self.score <= stopping_criterium:
                    logger.info(f"Sampler ended in {ii} iterations")
                    # Save model if converged
                    self.save_model()
                    break

        self.writer.close()
        return self.fit_points_x, self.fit_points_y

    def save_model(self):
        save_name = self.run_name + "_{}_iters.pkl".format(self.iterations)
        save_path = os.path.join(self.save_dir, save_name)

        with open(save_path, "wb") as file:
            pickle.dump(self.surrogate_model, file)


# -----------------------------------------------------
# SUPPORTING FUNCTIONS
# -----------------------------------------------------
def best_new_points(
    points_x: np.ndarray,
    points_y: np.ndarray,
    domain_lower_bound: np.ndarray,
    domain_upper_bound: np.ndarray,
    n_point_last_iter: int,
    n_new_point: int = 1,
    ignore_far_neighborhoods: Optional[bool] = True,
    ignore_old_neighborhoods: Optional[bool] = True,
):
    # shape the input if not in the right shape
    n_dim = len(domain_lower_bound)
    points_x = points_x.reshape((-1, n_dim))
    points_y = points_y.reshape((-1, 1))

    # Find the best neighborhoods for each srow of `points_x`
    (
        best_neighborhood_scores,
        best_neighborhood_idxs,
        all_neighborhood_scores,
        all_neighbor_point_idxs_combinations,
    ) = best_neighborhoods(
        points_x,
        n_point_last_iter,
        ignore_far_neighborhoods=ignore_far_neighborhoods,
        ignore_old_neighborhoods=ignore_old_neighborhoods,
    )

    # Find the `n_new_point_per_step` best next/new point(s)
    idx = all_neighbor_point_idxs_combinations[best_neighborhood_idxs]
    all_best_neighbor_points_x = points_x[idx, :]
    all_best_neighbor_points_y = points_y[idx, :]

    new_points_x = best_new_points_with_neighbors(
        reference_points_x=points_x,
        reference_points_y=points_y,
        all_neighbor_points_x=all_best_neighbor_points_x,
        all_neighbor_points_y=all_best_neighbor_points_y,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        n_next_point=n_new_point,
    )

    return new_points_x


def best_new_points_with_neighbors(
    reference_points_x: np.ndarray,
    reference_points_y: np.ndarray,
    all_neighbor_points_x: np.ndarray,
    all_neighbor_points_y: np.ndarray,
    domain_lower_bound: np.ndarray,
    domain_upper_bound: np.ndarray,
    n_next_point: int,
):
    """Find the new/next reference point(s) where the target function will be evaluated.

    TODO: reuse the random point used here for other calls of this function (
        `best_next_points`)
    """
    n_dim = reference_points_x.shape[1]

    nonlinearity_measures = lola_score(
        all_neighbor_points_x=all_neighbor_points_x,
        all_neighbor_points_y=all_neighbor_points_y,
        reference_points_x=reference_points_x,
        reference_points_y=reference_points_y,
    )
    (
        relative_volumes,
        random_points,
        distance_mx,
        closest_indicator_mx,
    ) = voronoi_volume_estimate(
        points=reference_points_x,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
    )
    hybrid_score = lola_voronoi_score(nonlinearity_measures, relative_volumes)

    # TODO: check np.partition as an alternative
    idxs_new_neighbor = np.argsort(-hybrid_score)[:n_next_point]

    new_reference_points_x = np.empty((n_next_point, n_dim))
    for ii, idx_new_neighbor in enumerate(idxs_new_neighbor):
        # all the distances to reference point whose neighborhood will get a new
        # point where the target function is evaluated
        distances_in_neighbor = distance_mx[:, idx_new_neighbor]

        # consider only those points that are within the voronoi cell of the
        # reference point
        idx_mask_in_neighbor = closest_indicator_mx[:, idx_new_neighbor]

        # to avoid selecting points that are not in the neighborhood
        distances_in_neighbor[~idx_mask_in_neighbor] = -1

        # largest distance within the same voronoi cell
        idx_max_distance = np.argmax(distances_in_neighbor)
        new_reference_point_x = random_points[idx_max_distance]
        new_reference_points_x[ii, :] = new_reference_point_x

    return new_reference_points_x


def best_neighborhoods(
    points_x: np.ndarray,
    n_point_last_iter: int,
    ignore_far_neighborhoods: Optional[bool] = True,
    ignore_old_neighborhoods: Optional[bool] = True,
):
    """Find the best neighborhood for each row of `points_x`. This function is expected
    to be used with the initial sample of `points_x`, when no best neighborhoods yet
    available from preceding iteration steps. This function exists to separate numba
    compatible code from non-compatible one."""

    # n-choose-k, all possible neighbor combinations, the elements of the matrix are
    # `points_x` indices
    n_point, n_dim = points_x.shape
    n_neighbor = 2 * n_dim

    all_neighbor_point_idxs_combinations = np.array(
        list(itertools.combinations(np.arange(n_point), n_neighbor))
    )
    (
        best_neighborhood_scores,
        best_neighborhood_idxs,
        all_neighborhood_scores,
    ) = best_neighborhoods_numba(
        points_x,
        all_neighbor_point_idxs_combinations,
        n_point_last_iter,
        ignore_far_neighborhoods=ignore_far_neighborhoods,
        ignore_old_neighborhoods=ignore_old_neighborhoods,
    )
    return (
        best_neighborhood_scores,
        best_neighborhood_idxs,
        all_neighborhood_scores,
        all_neighbor_point_idxs_combinations,
    )


def best_neighborhoods_numba(
    points_x: np.ndarray,
    all_neighbor_point_idxs_combinations: np.ndarray,
    n_point_last_iter: int,
    ignore_far_neighborhoods: Optional[bool] = True,
    ignore_old_neighborhoods: Optional[bool] = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Find the best neighborhood for each row of `points_x`. This function is expected
    to be used with the initial sample of `points_x`, when no best neighborhoods yet
    available from preceding iteration steps.

    Args:
        points_x: n_point x n_dim.
        all_neighbor_point_idxs_combinations:
        n_point_last_iter:
        ignore_far_neighborhoods:
        ignore_old_neighborhoods:
    Returns:

    """

    # TODO: It should be possible to perform the calculation of
    # all neighbourhoods after discarding the points that are
    # too far away. This will reduce storage requirements
    # and the computational cost of obtaining all combinations.
    n_point, n_dim = points_x.shape
    n_neighborhood = all_neighbor_point_idxs_combinations.shape[0]

    # Number of new points
    n_new_points = points_x.shape[0] - n_point_last_iter

    # this matrix contains neighborhoods that contain the reference point as well, these
    # will be assigned -1 and the rest will get a neighborhood score
    # each column belong to one point, same order as points_x (reference_point)
    all_neighborhood_scores = -np.ones((n_neighborhood, n_point))

    # TODO: both loops are completely independent hence parallelizable
    # for ii in nb.prange(n_point):
    # This loop is executed for each point in the domain
    for ii in range(n_point):
        reference_point_x = np.expand_dims(points_x[ii], 0)
        # consider only the valid neighborhoods: the ones that do not contain
        # `reference_point_x`
        idx_valid_neighborhoods = np.where(
            np_all(all_neighbor_point_idxs_combinations != ii, axis=1)
        )[0]

        if ignore_old_neighborhoods:
            # Find index of old points
            arr_valid_neighbourhoods = all_neighbor_point_idxs_combinations[
                idx_valid_neighborhoods
            ]
            arr_new_point_no = np.arange(
                n_point_last_iter, n_point_last_iter + n_new_points
            )

            # Remove entries from valid neighborhoods that
            # do not contain a new point
            arr_valid_neighbourhoods_mask = np.repeat(
                False, len(arr_valid_neighbourhoods)
            )
            for _idx_pt, pt in enumerate(arr_new_point_no):

                # Mask valid neighbourhoods (i.e. neighbourhoods that contain
                # the current new point). Terminate early if all elements
                # in mask array are true.
                arr_valid_neighbourhoods_mask = np.bitwise_or(
                    arr_valid_neighbourhoods_mask,
                    np_any(arr_valid_neighbourhoods == pt, axis=1),
                )
                if np.all(arr_valid_neighbourhoods_mask):
                    break
            arr_valid_neighbourhoods = arr_valid_neighbourhoods[
                arr_valid_neighbourhoods_mask
            ]
            idx_valid_neighborhoods = np.where(arr_valid_neighbourhoods_mask)[0]

        # consider only the new neighborhoods that are not too far away.
        # neighborhoods that contain a point farther away the median Euclidean distance
        # are dismissed.
        # section [5] of the paper
        if np.where(idx_valid_neighborhoods)[0].size:
            if ignore_far_neighborhoods:
                valid_pts_combinations = all_neighbor_point_idxs_combinations[
                    idx_valid_neighborhoods
                ]
                all_neighbor_points_x = points_x[valid_pts_combinations, :]
                dists = np.linalg.norm(
                    all_neighbor_points_x - reference_point_x, axis=-1
                )
                med = np.median(dists)
                idx_valid_neighborhoods = idx_valid_neighborhoods[
                    np.all(dists <= med, axis=-1)
                ]

        # compute the neighbourhood_score (`ns`) for each neighborhood
        #
        if np.where(idx_valid_neighborhoods)[0].size:
            for jj in range(len(idx_valid_neighborhoods)):
                idx_valid_neighborhood = idx_valid_neighborhoods[jj]
                neighbor_points_x = points_x[
                    all_neighbor_point_idxs_combinations[idx_valid_neighborhood], :
                ]
                # this is a time consuming function
                ns = neighborhood_score(
                    neighbor_points_x=neighbor_points_x,
                    reference_point_x=reference_point_x,
                )[0]
                all_neighborhood_scores[idx_valid_neighborhood, ii] = ns

    # best neighborhoods, the order is the same as in `points_x`
    # best_neighborhood_idxs = np_argmax(all_neighborhood_scores, axis=0)
    best_neighborhood_idxs = np.argmax(all_neighborhood_scores, axis=0)
    flat_idxs = n_point * best_neighborhood_idxs + np.arange(n_point)
    # best_neighborhood_scores = all_neighborhood_scores.ravel()[
    #     flat_idxs.astype(nb.int_)
    # ]
    best_neighborhood_scores = all_neighborhood_scores.ravel()[flat_idxs.astype(np.int)]
    return (
        best_neighborhood_scores,
        best_neighborhood_idxs,
        all_neighborhood_scores,
    )


def lola_voronoi_score(
    nonlinearity_measures: np.ndarray, relative_volumes: np.ndarray
) -> np.ndarray:
    """Eq.(5.1) of [1]."""
    return relative_volumes + nonlinearity_measures / np.sum(nonlinearity_measures)


def neighborhood_score(
    neighbor_points_x: np.ndarray, reference_point_x: np.ndarray
) -> Tuple[float, float]:
    """

    Args:
        neighbor_points_x:
            n_point x n_dim.
        reference_point_x:
            1 x n_dim

    Returns:
        Neighborhood score.
    """
    # shapes are not checked!
    # shape the input if not in the right shape, TODO: is this really needed?
    reference_point_x = reference_point_x.reshape((1, -1))
    n_dim = reference_point_x.shape[1]
    # neighbor_points_x = neighbor_points_x.reshape((-1, n_dim))

    # cohesion, Eq(4.3) of [1]
    cohesion = np.mean(
        np.sqrt(np.sum((neighbor_points_x - reference_point_x) ** 2, axis=1))
    )

    # cross-polytope ratio
    if n_dim == 1:
        pr1 = neighbor_points_x[0, 0]
        pr2 = neighbor_points_x[1, 0]
        # Eq(4.6) of [1]
        cross_polytope_ratio = 1 - np.abs(pr1 + pr2) / (
            np.abs(pr1) + np.abs(pr2) + np.abs(pr1 - pr2)
        )
    else:
        # Eq(4.4) of [1]
        # smallest neighbor distance for each neighbor
        # neighbor_distances = squareform(pdist(neighbor_points_x, metric="euclidean"))
        # np.fill_diagonal(neighbor_distances, np.inf)
        # min_neighbor_distances = np.min(neighbor_distances, axis=1)
        #  neighbor_distances = pdist_full_matrix(neighbor_points_x)

        # This should be an equivalent alternative to squareform + pdist
        # that also works with numba. Based on this stackoverflow answer:
        # https://stackoverflow.com/questions/60839615/
        # efficiently-calculating-a-euclidean-dist-matrix-in-numpy
        # TODO: Check
        # TODO: would be good to avoid np.max
        # neighbor_distances = squareform(pdist(neighbor_points_x, metric="euclidean"))

        neighbor_distances = euclidean_distance(neighbor_points_x)
        np.fill_diagonal(neighbor_distances, np.max(neighbor_distances))
        # min_neighbor_distances = np_min(neighbor_distances, axis=1)
        min_neighbor_distances = np_min(neighbor_distances, axis=1)
        adhesion = np.mean(min_neighbor_distances)
        # Eq(4.5) of [1]
        cross_polytope_ratio = adhesion / (np.sqrt(2) * cohesion)

    # Eq(4.7) of [1]
    return cross_polytope_ratio / cohesion, cross_polytope_ratio


def lola_score(
    all_neighbor_points_x: np.ndarray,
    all_neighbor_points_y: np.ndarray,
    reference_points_x: np.ndarray,
    reference_points_y: np.ndarray,
) -> np.ndarray:
    """Non-linearity measure for each point and its neighbor. Measures how much a
    neighborhood deviates from a hyperplane."""
    n_reference_point = len(reference_points_y)
    es = np.empty(n_reference_point)

    for (
        ii,
        (
            neighbor_points_x,
            neighbor_points_y,
            reference_point_x,
            reference_point_y,
        ),
    ) in enumerate(
        zip(
            all_neighbor_points_x,
            all_neighbor_points_y,
            reference_points_x,
            reference_points_y,
        )
    ):

        reference_point_gradient = gradient_estimate(
            reference_point_x=reference_point_x,
            reference_point_y=reference_point_y,
            neighbor_points_x=neighbor_points_x,
            neighbor_points_y=neighbor_points_y,
        )[0]
        es[ii] = nonlinearity_measure(
            reference_point_x=reference_point_x,
            reference_point_y=reference_point_y,
            reference_point_gradient=reference_point_gradient,
            neighbor_points_x=neighbor_points_x,
            neighbor_points_y=neighbor_points_y,
        )

    return es


def nonlinearity_measure(
    reference_point_x: np.ndarray,
    reference_point_y: float,
    reference_point_gradient: np.ndarray,
    neighbor_points_x: np.ndarray,
    neighbor_points_y: np.ndarray,
) -> float:
    # Eq.(4.9) of [1]
    e = np.sum(
        np.abs(
            neighbor_points_y
            - (
                reference_point_y
                + reference_point_gradient * (neighbor_points_x - reference_point_x)
            )
        )
    )
    return float(e)


def voronoi_volume_estimate(
    points: np.ndarray,
    domain_lower_bound: np.ndarray,
    domain_upper_bound: np.ndarray,
    n_simulation: int = None,
    random_points: np.ndarray = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Estimate the relative volume of the Voronoi tessellation of `points` bounded by
    `domain_lower_bound` and `domain_upper_bound`.

    The algorithm is described in section 3 of [1].

    Args:
        points:
            n_point x n_dim.
        domain_lower_bound:
            n_dim.
        domain_upper_bound:
            n_dim
        n_simulation:
            Number of random points used to estimate the relative volumes. If
            `random_points` is provided then this argument is ignored.
        random_points:

    Returns:
        Relative volumes in the same order as `points`.
    """
    # dimensions are not checked
    if n_simulation is None:
        n_simulation = 1000 * points.shape[0]

    if random_points is None:
        n_dim = len(domain_lower_bound)
        random_points = domain_lower_bound + np.random.rand(n_simulation, n_dim) * (
            domain_upper_bound - domain_lower_bound
        )

    # all relevant distances, n_simulation x n_point
    distance_mx = cdist(random_points, points, metric="euclidean")

    # index of the closest `point` to each `random_point`
    col_idx_min = np.argmin(distance_mx, axis=1)
    row_idx_min = np.arange(distance_mx.shape[0])

    # indicator matrix to count the number of `random_points` closest to each `points`
    closest_indicator_mx = np.zeros(distance_mx.shape, dtype=bool)
    # place a one (True) in the indicator matrix if the element (distance) is a closest
    # distance
    closest_indicator_mx[row_idx_min, col_idx_min] = True

    # count the closest `random_points` for each `points`
    closest_counts = np.sum(closest_indicator_mx, axis=0)

    # relative volume estimate
    relative_volumes = closest_counts / n_simulation

    return relative_volumes, random_points, distance_mx, closest_indicator_mx


def gradient_estimate(
    reference_point_x: np.ndarray,
    reference_point_y: float,
    neighbor_points_x: np.ndarray,
    neighbor_points_y: np.ndarray,
) -> np.ndarray:
    """
    Estimate the gradient at `reference_point` by fitting a hyperplane to
    `neighbor_points` in a least-square sense. A hyperplane that goes exactly
    through the `reference_point`.

    We think that the there is a mistake in Eq.(4.8) of [1], the right hand side
    should be `f(p_neighbor) - f(p_reference)`, or on the left hand side the
    `-p_reference` should be dropped if the formulation is in line with this:
    "Without loss of generality, we assume that pr lies in the origin."
    section 4.2.1 of [1].

    Args:
        reference_point_x:
            1 x n_dim.
        reference_point_y:
        neighbor_points_x:
            n_neighbor x n_dim.
        neighbor_points_y:
            n_neighbor x 1.

    Returns:
        Gradient estimate, 1 x n_dim.
    """
    # shape the input if not in the right shape, TODO: is this really needed?
    reference_point_x = reference_point_x.reshape((1, -1))
    n_dim = reference_point_x.shape[1]
    neighbor_points_x = neighbor_points_x.reshape((-1, n_dim))
    neighbor_points_y = neighbor_points_y.reshape((-1, 1))

    # to ensure that we hyperplane goes through `reference_point`
    neighbor_points_x_diff = neighbor_points_x - reference_point_x
    neighbor_points_y_diff = neighbor_points_y - reference_point_y

    # least-square fit of the hyperplane, the gradient is the hyperplane coefficients
    gradient = np.linalg.lstsq(
        neighbor_points_x_diff, neighbor_points_y_diff, rcond=None
    )[0].reshape((1, n_dim))

    return gradient



"""Test whether the surrogate model is converging to the target function."""
import matplotlib.pyplot as plt
import numpy as np

from harlow.sampling import LolaVoronoi
from harlow.surrogating import VanillaGaussianProcess
from harlow.utils.test_functions import forrester_1d, peaks_2d
from tests.integration_tests.utils import plot_1d_lola_voronoi


def test_sine_1d():
    n_new_points_per_iteration = [1, 2]
    n_initial_point = 10

    domain_lower_bound = np.array([-3])
    domain_upper_bound = np.array([3])
    plot_fig = True

    def target_function(x: np.ndarray):
        return np.sin(x)

    for n_new_point_per_iteration in n_new_points_per_iteration:

        n_iter = int(10 / n_new_point_per_iteration)

        surrogate_model = VanillaGaussianProcess()

        # ............................
        # Surrogating
        # ............................
        lv = LolaVoronoi(
            target_function=target_function,
            surrogate_model=surrogate_model,
            domain_lower_bound=domain_lower_bound,
            domain_upper_bound=domain_upper_bound,
        )
        lv.sample(
            max_n_iterations=n_iter,
            n_initial_points=n_initial_point,
            n_new_points_per_iteration=n_new_point_per_iteration,
        )

        # ............................
        # Check accuracy
        # ............................
        xx = np.linspace(domain_lower_bound, domain_upper_bound, 100)
        yy_tf = target_function(xx).ravel()
        yy_sm = lv.surrogate_model.predict(xx).ravel()

        np.testing.assert_allclose(yy_tf, yy_sm, atol=1e-1)

        # ............................
        # Plot
        # ............................
        if plot_fig:
            plot_1d_lola_voronoi(
                lv,
                n_initial_point=n_initial_point,
                n_new_point_per_iteration=n_new_point_per_iteration,
            )


def test_forrester_1d():
    n_new_point_per_iteration = 1
    n_initial_point = 10
    n_iter = 10

    domain_lower_bound = np.array([0])
    domain_upper_bound = np.array([1])
    plot_fig = True

    def target_function(x: np.ndarray):
        return forrester_1d(x)

    surrogate_model = VanillaGaussianProcess()

    # ............................
    # Surrogating
    # ............................
    lv = LolaVoronoi(
        target_function=target_function,
        surrogate_model=surrogate_model,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
    )
    lv.sample(
        max_n_iterations=n_iter,
        n_initial_points=n_initial_point,
        n_new_points_per_iteration=n_new_point_per_iteration,
    )

    # ............................
    # Check accuracy
    # ............................
    xx = np.linspace(domain_lower_bound, domain_upper_bound, 100)
    yy_tf = target_function(xx).ravel()
    yy_sm = lv.surrogate_model.predict(xx).ravel()

    np.testing.assert_allclose(yy_tf, yy_sm, atol=1)

    # ............................
    # Plot
    # ............................
    if plot_fig:
        plot_1d_lola_voronoi(
            lv,
            n_initial_point=n_initial_point,
            n_new_point_per_iteration=n_new_point_per_iteration,
        )


def test_peaks_2d():
    """Work in progress."""
    # section 6.2.3 of [1]
    n_iter = 6
    n_new_point_per_iteration = 2
    n_initial_point = 10

    domain_lower_bound = np.array([-5, -5])
    domain_upper_bound = np.array([5, 5])
    plot_fig = True

    def target_function(x: np.ndarray):
        x = np.atleast_2d(x)
        return peaks_2d(x).reshape(-1, 1)

    surrogate_model = VanillaGaussianProcess()

    # ............................
    # Surrogating
    # ............................
    lv = LolaVoronoi(
        target_function=target_function,
        surrogate_model=surrogate_model,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
    )
    lv.sample(
        max_n_iterations=n_iter,
        n_initial_points=n_initial_point,
        n_new_points_per_iteration=n_new_point_per_iteration,
    )

    # ............................
    # Check accuracy
    # ............................
    n_grid = 50
    x1_vec = np.linspace(domain_lower_bound[0], domain_upper_bound[0], n_grid)
    x2_vec = np.linspace(domain_lower_bound[1], domain_upper_bound[1], n_grid)
    x1_mx, x2_mx = np.meshgrid(x1_vec, x2_vec)
    x12_vec = np.vstack((x1_mx.ravel(), x2_mx.ravel())).T

    yy_tf = target_function(x12_vec).reshape((n_grid, n_grid))
    yy_sm = lv.surrogate_model.predict(x12_vec).reshape((n_grid, n_grid))

    # np.testing.assert_allclose(yy_tf, yy_sm, atol=1e-1)

    # ............................
    # Plot
    # ............................
    if plot_fig:
        fig, axs = plt.subplots(1, 2, sharex="all", sharey="all")
        ax1 = axs[0]
        ax2 = axs[1]

        ax1.contourf(x1_mx, x2_mx, yy_tf)
        ax1.set_xlabel("$x_1$")
        ax1.set_ylabel("$x_2$")
        ax1.set_title("Target function")

        cs = ax2.contourf(x1_mx, x2_mx, yy_sm)
        ax2.scatter(
            lv.fit_points_x[:, 0], lv.fit_points_x[:, 1], color="red", alpha=0.5
        )
        ax2.set_xlabel("$x_1$")
        ax2.set_ylabel("$x_2$")
        ax2.set_title("Surrogate function")
        # TODO: apply the same color range to both contourf plots
        plt.colorbar(cs, ax=axs)


test_sine_1d()
test_forrester_1d()
test_peaks_2d()



"""
IJsselbridge twin girder Euler-Bernoulli beam model.

This modified model does not have a Julia dependency, but instead
loads the pre-computed stiffness matrix.
"""

import pickle

# ============================================================================
# IMPORTS
# ============================================================================
from copy import copy, deepcopy
from pathlib import Path

import numpy as np
from numba import jit
from scipy.linalg import cho_factor, cho_solve

from harlow.utils.examples.model_data.truck_data import *  # noqa: F403


# ============================================================================
# FEM SOLVE FUNCTIONS
# ============================================================================
def lateral_load_func(z, c):
    z = np.asarray(z)

    # If c is a single value, assumes its a linear coefficient. otherwise its
    # assumed to be the n coefficients of a polynomial
    c = np.atleast_1d(c)
    if (not np.shape(c)) or (len(c) == 1):
        if not np.isscalar(c):
            return c[0] * (z - 2.85) + 0.5
        else:
            return c * (z - 2.85) + 0.5
    else:
        return np.polyval(c, z - 2.85)


@jit(nopython=True)
def find_nearest(array, values):
    # By default in case of multiple equal values the first is returned
    # (see numpy documentation for argmin())
    array = np.asarray(array)
    values = np.atleast_1d(np.asarray(values))
    idx = np.zeros(len(values), dtype=np.int32)
    for i, val in enumerate(values):
        idx[i] = np.abs(array - val).argmin()
    return idx


def find_all(array, values):
    array = np.asarray(array)
    values = np.atleast_1d(np.asarray(values))
    idx = []
    for _i, val in enumerate(values):
        idx.append(np.where(array == val))
    return np.ravel(idx)


# @jit(nopython = True)
def betti(nnode, cho_factor, W_sensor, hinge_node, hinge_left_node, fs, idx_keep):

    u = np.zeros(2 * nnode)

    # Note: Given equal Ks and fs, both the cholesky and conventional solve
    # will give slightly different results than the corresponding Julia
    # solve: us = Ks \ fs.
    us = cho_solve(cho_factor, fs, check_finite=False)

    u[idx_keep] = us
    r_hinge = u[hinge_left_node * 2 + 1] - u[hinge_node * 2 + 1]

    # scale to unit hinge rotation
    m_b_1 = 1 / r_hinge * u[0::4]
    m_b_2 = 1 / r_hinge * u[2::4]

    return np.asarray([m_b_1, m_b_2]) / W_sensor


def solve_inf_line(m_b, axle_pos_x, node_xs, loads):
    stress_1 = loads[:] * np.interp(axle_pos_x[:], node_xs, m_b[0])
    stress_2 = loads[:] * np.interp(axle_pos_x[:], node_xs, m_b[1])
    return np.sum(stress_1, axis=1), np.sum(stress_2, axis=1)


# ====================================================================
# INITIALIZATION
# ====================================================================
class IJssel_bridge_model:
    def __init__(
        self,
        sensor_name,
        E,
        max_elem_length,
        truck_load="TNO",
    ):
        # ====================================================================
        # INITIALIZATION
        # ====================================================================
        self.sensor_name = sensor_name
        self.max_elem_length = max_elem_length
        self.truck_load = truck_load
        self.E = E

        # Load model data
        fpath = Path(__file__).resolve().parent
        fname = f"model_dict#max_elem_length={self.max_elem_length}#E={self.E}.dat"

        with open(str(fpath / "model_data" / fname), "rb") as handle:
            model_properties = pickle.load(handle)

        # Get sensor names and positions
        self.sensor_names = model_properties["sensor_names"]
        self.sensor_positions = model_properties["sensor_positions"]

        # Get sensor index and position for this model
        idx_sensor = model_properties["sensor_names"].index(self.sensor_name)
        self.sensor_position = model_properties["sensor_positions"][idx_sensor]

        # Check that the requested sensor exists
        if self.sensor_name not in self.sensor_names:
            raise ValueError(
                f"Specified sensor name {self.sensor_name} not found in"
                f" {self.sensor_names}"
            )

        # Keep model properties for requested sensor
        self.model_properties = model_properties[self.sensor_name]

        # ====================================================================
        # STRUCTURE PROPERTIES FROM JULIA MODEL
        # ====================================================================

        self.hinge_left_pos = self.model_properties["hinge_left_pos"]
        self.support_xs = self.model_properties["support_xs"]
        self.node_xs = self.model_properties["node_xs"]
        self.elem_c_ys = self.model_properties["elem_c_ys"]
        self.elem_h_ys = self.model_properties["elem_h_ys"]
        self.elem_I_zs = self.model_properties["elem_I_zs"]
        self.nodes = self.model_properties["nodes"]

        # Element stiffness
        self.elem_EIs = self.elem_I_zs * self.E
        self.W_bot_temp = self.elem_I_zs / (self.elem_h_ys - self.elem_c_ys)
        self.elem_W_bottom = np.repeat(self.W_bot_temp, 2)
        self.nelems = len(self.elem_EIs)

        # NOTE: This gives the closest node.
        # It does not ensure that sensor_pos == node_pos
        self.sensor_node = find_nearest(self.nodes[:, 0], self.sensor_position)[0]
        self.hinge_node = self.sensor_node
        self.hinge_left_node = find_nearest(self.nodes[:, 0], self.hinge_left_pos)[0]
        self.W_sensor = self.elem_W_bottom[self.sensor_node]

        # Load stiffness matrix, nodes, spring positions
        self.Ks0 = self.model_properties["Ks0"]
        self.idx_keep = self.model_properties["idx_keep"]
        self.lin_idx_springs = self.model_properties["lin_idx_springs"]

        # Convert the flat, 1-indexed indices from Julia into numpy indices
        K_shape = np.shape(self.Ks0)
        lin_idx_spring_off_diag = np.asarray(self.lin_idx_springs["off_diag"]) - 1
        lin_idx_spring_diag = np.asarray(self.lin_idx_springs["diag"]) - 1
        self.idx_spring_off_diag = np.unravel_index(lin_idx_spring_off_diag, K_shape)
        self.idx_spring_diag = np.unravel_index(lin_idx_spring_diag, K_shape)

        # TODO: Use sparse matrices for Ks to improve performance.
        self.nnode = np.shape(self.nodes)[0]
        self.ndof = np.shape(self.Ks0)[0]
        self.support_nodes = find_all(self.nodes[:, 0], self.support_xs)

        # Boolean masking for applying rotational stiffnesses to Ks
        self.Ks_stiff_mask = np.asarray(np.zeros(2 * self.nnode), dtype=bool)
        self.Ks_stiff_mask[self.support_nodes * 2 + 1] = 1
        self.Ks_stiff_mask = self.Ks_stiff_mask[self.idx_keep]

        # Copy and factorize Ks. Ks and K are related as follows:
        # # Ks = K[idx_keep][:, idx_keep]
        self.Ks = deepcopy(self.Ks0)
        self.Ks_factor = cho_factor(self.Ks)

        # Precalculate unit load pair for betti theorem
        f = np.zeros(self.nnode * 2)
        f[self.hinge_left_node * 2 + 1] = 1e6
        f[self.hinge_node * 2 + 1] = -1e6
        self.fs = f[self.idx_keep]

        # Precalculate unit influence line
        self.unit_il = betti(
            self.nnode,
            self.Ks_factor,
            self.W_sensor,
            self.sensor_node,
            self.hinge_left_node,
            self.fs,
            self.idx_keep,
        )

        # ====================================================================
        # TRUCK DATA
        #
        # The truck data corresponding to 'TNO' or 'fugro' measurements is
        # loaded. Default is 'TNO'.
        # ====================================================================

        # Wheel z positions
        z_truck_r_wheel_r = truck_data[truck_load]["right"]["z_wheel_r"]  # noqa: F405
        z_truck_r_wheel_l = truck_data[truck_load]["right"]["z_wheel_l"]  # noqa: F405
        z_truck_l_wheel_r = truck_data[truck_load]["left"]["z_wheel_r"]  # noqa: F405
        z_truck_l_wheel_l = truck_data[truck_load]["left"]["z_wheel_l"]  # noqa: F405

        # Forces
        right_truck_force = truck_data[truck_load]["right"]["force"]  # noqa: F405
        left_truck_force = truck_data[truck_load]["left"]["force"]  # noqa: F405

        # Axle distances
        right_truck_axle_dist_x = truck_data[truck_load]["right"][  # noqa: F405
            "axle_dist"
        ]
        left_truck_axle_dist_x = truck_data[truck_load]["left"][  # noqa: F405
            "axle_dist"
        ]

        # Create objects holding the parameter values for left, right and double
        # case. This is to avoid the if-else statements that were used previously.
        self.truck_z = {
            "left": [z_truck_l_wheel_r, z_truck_l_wheel_l],
            "right": [z_truck_r_wheel_r, z_truck_r_wheel_l],
            "double": [
                z_truck_r_wheel_r,
                z_truck_r_wheel_l,
                z_truck_l_wheel_r,
                z_truck_l_wheel_l,
            ],
        }
        self.truck_force = {
            "left": np.transpose(left_truck_force),
            "right": np.transpose(right_truck_force),
            "double": np.vstack(
                (np.transpose(right_truck_force), np.transpose(left_truck_force))
            ),
        }
        self.truck_axle_dist = {
            "left": np.tile(left_truck_axle_dist_x, (2, 1)),
            "right": np.tile(right_truck_axle_dist_x, (2, 1)),
            "double": np.vstack(
                (
                    np.tile(right_truck_axle_dist_x, (2, 1)),
                    np.tile(left_truck_axle_dist_x, (2, 1)),
                )
            ),
        }

        # ====================================================================
        # PRECALCULATE LOAD PATHS AND INFLUENCE LINES
        #
        # NOTES:
        #   * This is not optimized for speed. This should not be a problem
        #   since this only runs once on model initialization.
        # ====================================================================

        # Dictionaries to store precalculation output
        all_paths = ["right", "left", "double"]
        self._truck_load = dict.fromkeys(all_paths)
        self.axle_pos_x = dict.fromkeys(all_paths)
        self.loaded_nodes = dict.fromkeys(all_paths)
        self.loads = dict.fromkeys(all_paths)

        for _i, lane in enumerate(all_paths):

            # Assemble the vectors of loads and axle distances
            loads = np.atleast_2d(self.truck_force[lane])
            axle_dist_x = np.atleast_2d(self.truck_axle_dist[lane])
            ax_dist = np.cumsum(axle_dist_x[0, :])
            load_i = loads[0, :]

            ax_pos = []
            l_node = []
            load = []
            for j in range(len(self.node_xs)):
                # For each load position sum the corresponding moments from the il and
                # calculate the stress at the sensor position
                curr_pos = np.append(self.node_xs[j], self.node_xs[j] - ax_dist)
                ax_pos.append(curr_pos)
                l_node.append(find_nearest(self.node_xs, curr_pos))

                # Make load zero for points that have not fully entererd the st
                # ructure yet
                temp_load = copy(load_i)
                temp_load[curr_pos < self.node_xs[0]] = 0
                load.append(temp_load)

            # Append to global list
            self.axle_pos_x[lane] = ax_pos
            self.loaded_nodes[lane] = l_node
            self.loads[lane] = load
            self._truck_load[lane] = np.tile(
                solve_inf_line(
                    self.unit_il, self.axle_pos_x[lane], self.node_xs, self.loads[lane]
                ),
                (len(self.truck_z[lane]), 1),
            )

    # ========================================================================
    # DEFINE FUNCTIONS
    # ========================================================================

    def il_stress_truckload(self, c, lane: str, Kr=None, Kv=None):
        """
        Calculate influence line

        Args:
           c: Lateral load coefficient
           lane: String, can be "left" or "right"
           Kr: Vector with 12 elements containing the
           support rotational stiffnesses. If omitted, the
           last calculated rotational stiffness vector is used
           Kv: Scalar vertical spring stiffness
        """
        # Check if stiffnesses have been updated, if yes, update the il
        if (lane == "double") or (lane == "Double"):
            raise ValueError("Not implemented")

        if (Kr is not None) or (Kv is not None):

            # Add new stiffnesses to stiffness matrix
            self.Ks = copy(self.Ks0)
            self.Ks[self.Ks_stiff_mask, self.Ks_stiff_mask] += np.asarray(Kr)
            self.Ks[self.idx_spring_diag] += Kv
            self.Ks[self.idx_spring_off_diag] -= Kv

            # Factorize
            Ks_factor = cho_factor(self.Ks)

            # Calculate new influence line
            self.unit_il = betti(
                self.nnode,
                Ks_factor,
                self.W_sensor,
                self.sensor_node,
                self.hinge_left_node,
                self.fs,
                self.idx_keep,
            )
        # The returned stress arrays contain two stress influence lines for each
        # truck z value. lines [0::2] correspond to the right girder stress and
        # lines [1::2] correspond to the left girder stress
        self._truck_load[lane] = np.tile(
            solve_inf_line(
                self.unit_il, self.axle_pos_x[lane], self.node_xs, self.loads[lane]
            ),
            (len(self.truck_z[lane]), 1),
        )

        stress = np.zeros(len(self.node_xs))
        for i, z in enumerate(self.truck_z[lane]):
            stress += lateral_load_func(z, c) * self._truck_load[lane][2 * i, :]
            stress += lateral_load_func(z, -c) * self._truck_load[lane][2 * i + 1, :]

        return stress / 1000  # Return stress in MPa



"""Benchmarktest from Crombecq et al. (2011). Benchmarks LV and
Probabilistic sampling vs random sampling. """
# import matplotlib.pyplot as plt
# import numpy as np
#
# from harlow.lola_voronoi import LolaVoronoi
# from harlow.surrogate_model import GaussianProcess
# from tests.integration_tests.test_functions import shekel


def test_LV_sampling():
    pass


def test_probabilistic_sampling():
    pass


def test_random_sampling():
    pass



import numpy as np
from matplotlib import pyplot as plt

from harlow.sampling import LolaVoronoi


def plot_1d_lola_voronoi(
    lv: LolaVoronoi,
    n_initial_point: int,
    n_new_point_per_iteration: int,
    n_grid_point: int = 100,
):
    """Utility function for visualizing the results of surrogating 1D functions with
    LolaVoronoi."""
    n_iter = int((len(lv.fit_points_y) - n_initial_point) / n_new_point_per_iteration)
    xx = np.linspace(lv.domain_lower_bound, lv.domain_upper_bound, n_grid_point)
    yy_tf = lv.target_function(xx).ravel()
    yy_sm = lv.surrogate_model.predict(xx)

    adaptive_points_x = lv.fit_points_x[n_initial_point:]
    adaptive_points_y = lv.fit_points_y[n_initial_point:]

    fig, ax = plt.subplots()
    ax.plot(xx, yy_tf, label="target function")
    ax.plot(xx, yy_sm, "--", label="surrogate function")
    ax.scatter(
        lv.fit_points_x[:n_initial_point],
        lv.fit_points_y[:n_initial_point],
        label="initial points",
        alpha=0.5,
    )
    ax.scatter(adaptive_points_x, adaptive_points_y, label="adaptive points", alpha=0.5)
    for ii in range(n_iter):
        idx_start = ii * n_new_point_per_iteration
        idx_end = (ii + 1) * n_new_point_per_iteration
        xs = adaptive_points_x[idx_start:idx_end]
        ys = adaptive_points_y[idx_start:idx_end]

        for jj, (x, y) in enumerate(zip(xs, ys)):
            if len(xs) == 1:
                ax.text(x, y, f"${ii + 1}$")
            else:
                ax.text(x, y, f"${ii + 1}^{jj + 1}$")

    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.legend()
    return fig, ax



"""
Rectangularly bounded Voronoi diagram in 2D.

The code is largely based on: https://stackoverflow.com/a/33602171/4063376
It is changed to match the order of `filtered_regions` with the order of `towers`.
"""
import sys

import numpy as np
import scipy as sp
import scipy.spatial

eps = sys.float_info.epsilon


def in_box(towers, bounding_box):
    return np.logical_and(
        np.logical_and(
            bounding_box[0] <= towers[:, 0], towers[:, 0] <= bounding_box[1]
        ),
        np.logical_and(
            bounding_box[2] <= towers[:, 1], towers[:, 1] <= bounding_box[3]
        ),
    )


def bounded_voronoi_2d(towers: np.ndarray, bounding_box: list):
    # bounding_box: [x_min, x_max, y_min, y_max]
    # Select towers inside the bounding box
    i = in_box(towers, bounding_box)
    # Mirror points
    points_center = towers[i, :]
    points_left = np.copy(points_center)
    points_left[:, 0] = bounding_box[0] - (points_left[:, 0] - bounding_box[0])
    points_right = np.copy(points_center)
    points_right[:, 0] = bounding_box[1] + (bounding_box[1] - points_right[:, 0])
    points_down = np.copy(points_center)
    points_down[:, 1] = bounding_box[2] - (points_down[:, 1] - bounding_box[2])
    points_up = np.copy(points_center)
    points_up[:, 1] = bounding_box[3] + (bounding_box[3] - points_up[:, 1])
    points = np.append(
        points_center,
        np.append(
            np.append(points_left, points_right, axis=0),
            np.append(points_down, points_up, axis=0),
            axis=0,
        ),
        axis=0,
    )
    # Compute Voronoi
    vor = sp.spatial.Voronoi(points)
    # Filter regions
    regions = []
    idx_filtered_regions = []
    for ii, region in enumerate(vor.regions):
        flag = True
        for index in region:
            if index == -1:
                flag = False
                break
            else:
                x = vor.vertices[index, 0]
                y = vor.vertices[index, 1]
                if not (
                    bounding_box[0] - eps <= x <= bounding_box[1] + eps
                    and bounding_box[2] - eps <= y <= bounding_box[3] + eps
                ):
                    flag = False
                    break
        if region != [] and flag:
            regions.append(region)
            idx_filtered_regions.append(ii)

    # until we can get the order right..
    # unordered_regions = regions
    # unordered_point_regions = vor.point_region[: towers.shape[0]]
    # ordered_regions = [None] * len(regions)
    # for ii, unordered_point_region in enumerate(unordered_point_regions):
    #     idx = int(np.where(unordered_point_region == idx_filtered_regions)[0])
    #     ordered_regions[ii] = unordered_regions[idx]
    ordered_regions = regions

    vor.filtered_points = points_center
    vor.filtered_regions = ordered_regions
    return vor
